#!/usr/bin/env python2
# coding: utf-8

import unittest

from pykit import strutil
from pykit import dictutil
from pykit import ututil

dd = ututil.dd


def dd_lines(lines):
    dd()
    for l in lines:
        dd(l)


class TestStrutil(unittest.TestCase):

    def test_tokenize(self):

        base_cases = (
                ('', [''], ''),
                ('a', ['a'], ''),
                ('a b', ['a', 'b'], ''),
                (' a', ['', 'a'], ''),
                ('b ', ['b', ''], ''),
                ('abc def gh', ['abc', 'def', 'gh'], ''),
                ('"ab cd"', ['"ab cd"'], ''),
                ('"ab  cd"', ['"ab  cd"'], 'multiple space inside quotes'),
                ('"ab cd" ', ['"ab cd"', ''], ''),
                (' "ab cd"', ['', '"ab cd"'], ''),
                ('"ab cd" "x', ['"ab cd"'], 'discard incomplete'),
                ('"ab cd" "x y"', ['"ab cd"', '"x y"'], ''),

                ('foo "ab cd" "x y"', ['foo', '"ab cd"', '"x y"'], ''),
                ('foo "ab cd" "x', ['foo', '"ab cd"'], 'discard incomplete'),

                ('foo "\\"ab cd" "x', ['foo', '""ab cd"'], 'escape "'),
                ('foo "\\\\"ab cd" "x', ['foo', '"\\"ab', 'cd" "x'], 'escape \\'),
                ('foo "\\\\\\"ab cd" "x', ['foo', '"\\"ab cd"'], 'escape \\ "'),

                ('a \\"bc "d e" "f',  ['a', '"bc', '"d e"'], ''),
                ('a \\\\"bc "d e" "f',  ['a', '\\"bc "d', 'e" "f'], ''),
                ('a \\\\\\"bc "d e" "f',  ['a', '\\"bc', '"d e"'], ''),

                ('a "bc "d \\"f',  ['a', '"bc "d', '"f'], ''),
                ('a "bc "d \\\\"f',  ['a', '"bc "d'], ''),
                ('a "bc "d \\\\\\"f',  ['a', '"bc "d', '\\"f'], ''),

                ('\\"bc "d "f',  ['"bc', '"d "f'], ''),
                ('\\\\"bc "d "f',  ['\\"bc "d'], ''),
                ('\\\\\\"bc "d "f',  ['\\"bc', '"d "f'], ''),

                ('a "bc "d f\\"',  ['a', '"bc "d', 'f"'], ''),
                ('a "bc "d f\\\\"',  ['a', '"bc "d'], ''),
                ('a "bc "d f\\\\\\"',  ['a', '"bc "d', 'f\\"'], ''),
        )

        for _in, _out, _mes in base_cases:
            rst = strutil.tokenize(_in, sep=' ', preserve=True)
            self.assertEqual(_out, rst,
                             ('input: {_in}, output: {_out}, expected: {rst},'
                              ' message: {_mes}').format(
                                 _in=repr(_in),
                                 _out=repr(_out),
                                 rst=repr(rst),
                                 _mes=_mes
                             ))

        sep_cases = (
                ('',              None,    True,   []),
                (' a  b  c ',     None,    True,   ['a', 'b', 'c']),
                (' a  "b  c" ',   None,    True,   ['a', '"b  c"']),
                (' a  "b  c" ',   None,    False,  ['a', 'b  c']),
                ('a b c',         None,    True,   ['a', 'b', 'c']),
                ('"a b c"',       None,    True,   ['"a b c"']),
                ('"a b c"',       None,    False,  ['a b c']),
                ('a b"c d"',      None,    True,   ['a', 'b"c d"']),
                ('a b"c d"',      None,    False,  ['a', 'bc d']),
                ('a bcd',         'bc',    True,   ['a ', 'd']),
                ('a "bc" d',      'bc',    True,   ['a "bc" d']),
                ('a "bc" d',      'bc',    False,  ['a bc d']),
                ('abcd',          'abcd',  True,   ['', '']),
        )

        for line, sep, preserve, rst_expected in sep_cases:
            dd('in: ', line, sep)
            rst = strutil.tokenize(line, sep=sep, quote='"', preserve=preserve)
            dd('out: ', rst)
            self.assertEqual(rst, rst_expected)

        preserve_cases = (
                ('""',                   '"',    True,    ['""']),
                ('""',                   '"',    False,   ['']),
                ('abc xd efx gh',        'x',    True,    ['abc', 'xd efx', 'gh']),
                ('abc xd efx gh',        'x',    False,   ['abc', 'd ef', 'gh']),
                ('ab cxd efx gh',        'x',    True,    ['ab', 'cxd efx', 'gh']),
                ('ab cxd efx gh',        'x',    False,   ['ab', 'cd ef', 'gh']),
                ('ab cxd efxgh',         'x',    True,    ['ab', 'cxd efxgh']),
                ('ab cxd efxgh',         'x',    False,   ['ab', 'cd efgh']),
                ('ab cxd yey fx gh',     'xy',   True,    ['ab', 'cxd yey fx', 'gh']),
                ('ab cxd yey fx gh',     'xy',   False,   ['ab', 'cd yey f', 'gh']),
                ('ab cxd yey f gh',      'xy',   True,    ['ab']),
                ('ab cxd yey f gh',      'xy',   False,   ['ab']),
                ('ab cxd xex f gh',      'x',    True,    ['ab']),
                ('ab cxd xex f gh',      'x',    False,   ['ab']),
        )

        for line, quote, preserve, rst_expected in preserve_cases:
            dd('in: ', line, quote, preserve)
            rst = strutil.tokenize(line, sep=' ', quote=quote, preserve=preserve)
            dd('out: ', rst)
            self.assertEqual(rst, rst_expected)

    def test_common_prefix_invalid_arg(self):
        cases = (
            (1, []),
            ('a', 1),
            ('a', True),
            ('a', ('a',)),
            (('a', (),), ('a', 2,)),
        )

        for a, b in cases:
            dd('wrong type: ', repr(a), ' ', repr(b))
            self.assertRaises(TypeError, strutil.common_prefix, a, b)

    def test_common_prefix(self):
        cases = (
            ('abc',
             'abc',
             ),
            ('',
             '',
             '',
             ),
            ((),
             (),
             (),
             (),
             ),
            ('abc',
             'ab',
             'ab',
             ),
            ('ab',
             'abd',
             'ab',
             ),
            ('abc',
             'abd',
             'ab',
             ),
            ('abc',
             'def',
             '',
             ),
            ('abc',
             '',
             '',
             ),
            ('',
             'def',
             '',
             ),
            ('abc',
             'abd',
             'ag',
             'a',
             ),
            ('abc',
             'abd',
             'ag',
             'yz',
             '',
             ),
            ((1, 2,),
             (1, 3,),
             (1,),
             ),
            ((1, 2,),
             (2, 3,),
             (),
             ),
            ((1, 2, 'abc', ),
             (1, 2, 'abd', ),
             (1, 2, 'ab', ),
             ),
            ((1, 2, 'abc', ),
             (1, 2, 'xyz', ),
             (1, 2, ),
             ),
            ((1, 2, (5, 6), ),
             (1, 2, (5, 7), ),
             (1, 2, (5,), ),
             ),
            (('abc', '45',),
             ('abc', '46', 'xyz'),
             ('abc', '4',),
             ),
            (('abc', ('45', 'xyz'), 3),
             ('abc', ('45', 'xz'), 5),
             ('abc', ('45', 'x-'), 5),
             ('abc', ('45', 'x'),),
             ),
            (('abc', ('45', 'xyz'), 3),
             ('abc', ('45', 'xz'), 5),
             ('abc', ('x',),),
             ('abc',),
             ),
            ([1, 2, 3],
             [1, 2, 4],
             [1, 2],
            ),
        )

        for args in cases:
            expected = args[-1]
            args = args[:-1]

            dd('input: ', args, 'expected: ', expected)
            rst = strutil.common_prefix(*args)
            dd('rst: ', rst)

            self.assertEqual(expected, rst)

    def test_common_prefix_no_recursive(self):

        cases = (
            (('abc', ('45', 'xyz'), 3),
             ('abc', ('45', 'xz'), 5),
             ('abc', ('45', 'x-'), 5),
             ('abc',),
             ),
            ('abc',
             'abd',
             'ag',
             'a',
             ),
            ((1, 2, 'abc'),
             (1, 2, 'abd'),
             (1, 2),
             ),
        )

        for args in cases:
            expected = args[-1]
            args = args[:-1]

            dd('input: ', args, 'expected: ', expected)
            rst = strutil.common_prefix(*args, recursive=False)
            dd('rst: ', rst)

            self.assertEqual(expected, rst)


    def test_line_pad(self):

        cases = (
                (('', ''),
                 '',
                 ''
                 ),
                (('a', ''),
                 'a',
                 ''
                 ),
                (('a', ' '),
                 ' a',
                 ''
                 ),
                (('a', '   '),
                 '   a',
                 ''
                 ),
                (('a\nb', '   '),
                 '   a\n'
                 '   b',
                 ''
                 ),
                (('a\nbc', lambda line: 'x' * len(line)),
                 'xa\n'
                 'xxbc',
                 ''
                 ),
        )

        for _in, _out, _mes in cases:
            rst = strutil.line_pad(*_in)
            self.assertEqual(_out, rst,
                             ('input: {_in}, output: {_out}, expected: {rst},'
                              ' message: {_mes}').format(
                                 _in=repr(_in),
                                 _out=repr(_out),
                                 rst=repr(rst),
                                 _mes=_mes
                             ))

    def test_format_line(self):

        cases = (
                (([], '', ''),
                 '',
                 ''
                 ),
                ((['a'], '', ''),
                 'a',
                 ''
                 ),
                (([['a', 'bc']], '', ''),
                 ' a\n'
                 'bc',
                 'default alignment is to right'
                 ),
                (([['a', 'bc'], 'foo'], '', ''),
                 ' afoo\n'
                 'bc   ',
                 'no sep, add space to align'
                 ),
                (([['a', 'bc'], 'foo'], '|', ''),
                 ' a|foo\n'
                 'bc|   ',
                 'sep is "|"'
                 ),
                (([['a', 'bc'], 'foo', [1, 2, 333]], '|', ''),
                 ' a|foo|  1\n'
                 'bc|   |  2\n'
                 '  |   |333',
                 'number will be convert to str'
                 ),
                (([['a', 'bc'], 'foo', [1, 2, strutil.blue('xp')]], '|', ''),
                 ' a|foo| 1\n'
                 'bc|   | 2\n'
                 '  |   |' + str(strutil.blue('xp')),
                 'strutil.ColoredString instance is compatible'
                 ),
        )

        for _in, _out, _mes in cases:

            rst = strutil.format_line(*_in)

            dd("_in: " + str(_in))
            dd("rst:\n" + rst)

            self.assertEqual(_out, rst,
                             ('input: {_in}, output: {_out}, expected: {rst},'
                              ' message: {_mes}').format(
                                 _in=repr(_in),
                                 _out=repr(_out),
                                 rst=repr(rst),
                                 _mes=_mes
                             ))

    def test_struct_repr(self):

        inp = {
            1: 3,
            'x': {1: 4, 2: 5},
            'yyy': [1, 2, 3, 1000],
        }

        rst = strutil.struct_repr(inp)

        for l in rst:
            dd(repr(l))

        expected = [
            '  1 : 3     ',
            '  x : 1 : 4 ',
            '      2 : 5 ',
            'yyy : - 1   ',
            '      - 2   ',
            '      - 3   ',
            '      - 1000',
        ]

        self.assertEqual(expected, rst)

    def test_format_table(self):

        inp = [
            {'acl': {},
             'bucket': 'game1.read',
             'bucket_id': '1400000000000689036',
             'num_used': '0',
             'owner': 'game1',
             'space_used': '0',
             'ts': '1492091893065708032'},
            {'acl': {},
             'bucket': 'game2.read',
             'bucket_id': '1510000000000689037',
             'num_used': '0',
             'owner': 'game2',
             'space_used': '0',
             'ts': '1492091906629786880'},
            {'acl': {'imgx': ['READ', 'READ_ACP', 'WRITE', 'WRITE_ACP']},
             'bucket': 'imgx-test',
             'bucket_id': '1910000000000689048',
             'num_used': '0',
             'owner': 'imgx',
             'space_used': '0',
             'ts': '1492101189213795840'}]

        # default
        rst = strutil.format_table(inp)
        dd_lines(rst)

        # the last line is a '\n' splitted multi-row line
        expected = [
            'acl:               | bucket:    | bucket_id:          | num_used:  | owner:  | space_used:  | ts:                ',
            '{}                 | game1.read | 1400000000000689036 | 0          | game1   | 0            | 1492091893065708032',
            '{}                 | game2.read | 1510000000000689037 | 0          | game2   | 0            | 1492091906629786880',
            'imgx : - READ      | imgx-test  | 1910000000000689048 | 0          | imgx    | 0            | 1492101189213795840\n'
            '       - READ_ACP  |            |                     |            |         |              |                    \n'
            '       - WRITE     |            |                     |            |         |              |                    \n'
            '       - WRITE_ACP |            |                     |            |         |              |                    ',
        ]
        self.assertEqual(expected, rst)

        # specify key to render

        rst = strutil.format_table(inp, keys=[['bucket', 'B']])
        dd_lines(rst)
        expected = [
            'B:        ',
            'game1.read',
            'game2.read',
            'imgx-test ',
        ]
        self.assertEqual(expected, rst)

        # row_sep
        rst = strutil.format_table(inp, keys=['bucket'], row_sep='+')
        dd_lines(rst)
        expected = [
            'bucket:   ',
            '++++++++++',
            'game1.read',
            '++++++++++',
            'game2.read',
            '++++++++++',
            'imgx-test ',
        ]
        self.assertEqual(expected, rst)

        # sep
        rst = strutil.format_table(
            inp, keys=['bucket', 'bucket_id'], sep=' # ')
        dd_lines(rst)
        expected = [
            'bucket:    # bucket_id:         ',
            'game1.read # 1400000000000689036',
            'game2.read # 1510000000000689037',
            'imgx-test  # 1910000000000689048',
        ]
        self.assertEqual(expected, rst)

    def test_utf8str(self):
        cases = (
                ('', ''),
                ('1', '1'),
                (1, '1'),
                (u'我', '\xe6\x88\x91'),
                ('\xe6\x88\x91', '\xe6\x88\x91'),
        )

        for inp, expected in cases:

            rst = strutil.utf8str(inp)

            self.assertEqual(expected, rst)

    def test_break_line(self):
        case = [
            ('a quick brown fox jumps over the lazy dog',        13,    ['a quick brown', 'fox jumps', 'over the lazy', 'dog']),
            ('a quick\nbrown fox jumps over the lazy dog',       13,    ['a quick', 'brown fox', 'jumps over', 'the lazy dog']),
            ('a quick\rbrown fox jumps over the lazy dog',       13,    ['a quick', 'brown fox', 'jumps over', 'the lazy dog']),
            ('a quick brown fox jumps\r\nover the lazy dog',     13,    ['a quick brown', 'fox jumps', 'over the lazy', 'dog']),
            ('a quick\nbrown\rfox jumps\r\nover the lazy dog',   13,    ['a quick', 'brown', 'fox jumps', 'over the lazy', 'dog']),
            ('aquickbrownfoxjumpsoverthelazydog',                9,     ['aquickbrownfoxjumpsoverthelazydog']),
            ('aquickbro',                                        9,     ['aquickbro']),
            (' aquickbro',                                       9,     ['', 'aquickbro']),
            ('  aquickbro',                                      9,     [' ', 'aquickbro']),
            ('aquickbro ',                                       9,     ['aquickbro']),
            ('aquickbro  ',                                      9,     ['aquickbro', ' ']),
            ('aqu ick br',                                       9,     ['aqu ick', 'br']),
            ('aqu ick br',                                       9.34,  ['aqu ick', 'br']),
            ('apu   ick  br',                                    5,     ['apu  ', 'ick ', 'br']),
            ('aqu ick br',                                       0,     ['aqu', 'ick', 'br']),
            ('aqu ick br',                                       -1,    ['aqu', 'ick', 'br']),
            ('',                                                 2,     []),
            (' ',                                                2,     [' ']),
            ('  ',                                               2,     ['  ']),
            ('   ',                                              2,     ['  ']),
            ('    ',                                             2,     ['  ', ' ']),
        ]

        for linestr, width, expected in case:
            rst = strutil.break_line(linestr, width)
            dd('case: ', linestr, width, expected)
            dd('rst: ', rst)
            self.assertEqual(rst, expected)

        colored_string_cases = [
            ([('asd ', 'red'), ('fer ', 'blue'), ('fg', 'white')],
             8,
             [[('asd', 'red'), (' ', None), ('fer', 'blue')],  [('fg', 'white')]],
             'base test',
             ),

            ([('asd ', 'red'), ('fer ', 'blue'),  ('fg', 'white')],
             -1,
             [[('asd', 'red')], [('fer', 'blue')],  [('fg', 'white')]],
             'test width',
             ),

            ([('asd   ', 'red'), ('fer  ', 'blue'),  ('fg', 'white')],
             5,
             [[('asd', 'red'), (' ', None), (' ', None)],
              [('fer', 'blue'), (' ', None)], [('fg', 'white')]],
             'test consecutive blank spaces',
             ),

            ([('  asd', 'red'), (' fer', 'blue'), ('fg   ', 'white')],
             4,
             [[(' ', None)], [('asd', 'red')], [('fer', 'blue'), ('fg', 'white')],
              [(' ', None), (' ', None)]],
             'consecutive blank spaces in the end and beginning',
             ),

            ([('asd\n', 'red'), ('f er ', 'blue'),  ('fg', 'white')],
             5,
             [[('asd', 'red')], [('f', 'blue'), (' ', None), ('er', 'blue')], [('fg', 'white')]],
             'test line break',
             ),

            ([('', None)],
             2,
             [],
             'test empty string',
             ),
        ]

        for _in, width, expected, msg in colored_string_cases:
            dd('msg: ', msg)

            color_in = strutil.ColoredString('')
            for elt in _in:
                color_in += strutil.ColoredString(elt[0], elt[1])
            color_expected = []
            for l in expected:
                buf = strutil.ColoredString('')
                for elt in l:
                    buf += strutil.ColoredString(elt[0], elt[1])
                color_expected.append(buf)

            rst = strutil.break_line(color_in, width)

            dd('rst: ', rst)
            dd('expected: ', color_expected)
            self.assertEqual(rst, color_expected)


class TestColoredString(unittest.TestCase):

    cs = strutil.ColoredString

    def test_colorize_input(self):

        cases = (
                (0, 0),

                (0, 1),
                (0, 100),
                (1, 100),
                (50, 100),
                (100, 100),

                (0, -1),
                (0, -100),
                (1, -100),
                (50, -100),
                (100, -100),
        )

        print

        for v, total in cases:
            print strutil.colorize(v, total, str(v) + '/' + str(total))

    def test_show_all_colors(self):

        for c in range(256):
            if c % 16 == 0:
                print
            print self.cs('{0:>3}'.format(c), c),

    def test_named_color(self):

        print

        # named color shortcuts
        print strutil.blue('blue'),
        print strutil.cyan('cyan'),
        print strutil.green('green'),
        print strutil.yellow('yellow'),
        print strutil.red('red'),
        print strutil.purple('purple'),
        print strutil.white('white'),
        print

        print strutil.danger('danger'),
        print strutil.warn('warn'),
        print strutil.loaded('loaded'),
        print strutil.normal('normal'),
        print strutil.optimal('optimal'),
        print

    def test_pading(self):

        print
        for i in range(0, 100, 5):
            print strutil.colorize(i, 100),
        print

        for i in range(0, 100, 5):
            print strutil.colorize(i, -100),
        print

    def test_length(self):

        cases = (
                '',
                'string',
                '幾時何時',
                '\xf3',
        )

        for v in cases:
            self.assertEqual(len(self.cs(v, 'warn')), len(v))

    def test_add(self):

        # concat colored string with '+', like normal string
        s = self.cs('danger', 'danger') + self.cs('warn', 'warn')
        self.assertEqual(len('danger' + 'warn'), len(s))

        s += 'extra_string'
        self.assertEqual(len('danger' + 'warn' + 'extra_string'), len(s))

    def test_mul(self):

        # colored string can be duplicated with '*', like normal string
        s = (self.cs('danger', 'danger')
             + self.cs('warn', 'warn')
             + self.cs(' normal')) * 3

        slen = len('danger' + 'warn' + ' normal') * 3
        self.assertEqual(slen, len(s))

        s *= 10
        slen *= 10
        self.assertEqual(slen, len(s))

    def test_rerender(self):

        print

        # re-render strutil.ColoredString
        s = 'danger rerender to warn'

        c = self.cs(s, 'danger')
        print 'colorize with "danger":', c

        c = self.cs(c, 'warn')
        print 'colorize with "warn"  :', c

    def test_colored_prompt(self):

        s = '[colored prompt]# '

        prompt = self.cs(s, color='optimal', prompt=True)
        not_prompt = self.cs(s, color='optimal', prompt=False)

        self.assertEqual(str(prompt)[0], '\001')
        self.assertEqual(str(prompt)[-1], '\002')
        self.assertEqual(len(str(prompt)), len(str(not_prompt)) + 4)

        prompt += 'whatever'
        not_prompt += 'whatever'
        self.assertEqual(len(str(prompt)), len(str(not_prompt)) + 4)

        prompt *= 3
        not_prompt *= 3
        self.assertEqual(len(str(prompt)), len(str(not_prompt)) + 4*3)

    def test_split(self):
        color_cases = [
            ([('asd ', 'red'), ('fer ', 'blue'),  ('fg', 'white')],
             (' ', 0),
             [[('asd ', 'red'), ('fer ', 'blue'),  ('fg', 'white')]],
             'test maxsplit 0',
             ),

            ([('asd ', 'red'), ('fer ', 'blue'),  ('fg', 'white')],
             (' ', 1),
             [[('asd', 'red')], [('fer ', 'blue'),  ('fg', 'white')]],
             'test maxsplit 1',
             ),

            ([('asd ', 'red'), ('fer ', 'blue'),  ('fg', 'white')],
             (' ', -1),
             [[('asd', 'red')], [('fer', 'blue')],  [('fg', 'white')]],
             'test maxsplit -1',
             ),

            ([('asdx', 'red'), ('yferx', 'blue'),  ('yfg', 'white')],
             ('xy', -1),
             [[('asd', 'red')], [('fer', 'blue')], [('fg', 'white')]],
             'diff color separator',
             ),

            ([('asdx', 'red'), ('yferx', 'blue'),  ('yfg', 'white')],
             ('xy', 1),
             [[('asd', 'red')], [('ferx', 'blue'), ('yfg', 'white')]],
             'diff color separator and maxsplit',
             ),

            ([('asdx', 'red'), ('yferx', 'blue'),  ('yfg', 'white')],
             ('xyz', -1),
             [[('asdx', 'red'), ('yferx', 'blue'), ('yfg', 'white')]],
             'no separator in ColoredString',
             ),

            ([('  asd ', 'red'), (' fer ', 'blue'),  (' fg ', 'white')],
             (None, -1),
             [[('asd', 'red')], [('fer', 'blue')], [('fg', 'white')]],
             'separator is None',
             ),

            ([('  asd ', 'red'), (' fer ', 'blue'),  (' fg ', 'white')],
             (None, 1),
             [[('asd', 'red')], [('fer ', 'blue'), (' fg ', 'white')]],
             'separator is None and test maxsplit',
             ),

            ([('  ', 'red'), (' \r', 'blue'), ('\n  ', 'white')],
             [None, -1],
             [],
             'whitespace string and separator is None',
             ),

            ([('  ', 'red'), (' \r', 'blue'), ('\n  ', 'white')],
             [' ', -1],
             [[('', 'red')], [('', 'red')], [('', 'red')],
              [('\r', 'blue'), ('\n', 'white')], [('', 'red')], [('', 'red')]],
             'consecutive separator string',
             ),

            ([('', 'red')],
             [' ', 1],
             [[('', None)]],
             'blank string',
             ),
        ]

        for _in, args, expected, msg in color_cases:
            dd('msg: ', msg)

            expect_rsts = []
            for elts in expected:
                cs = strutil.ColoredString('')
                for elt in elts:
                    cs += strutil.ColoredString(elt[0], elt[1])
                expect_rsts.append(cs)

            color_in = strutil.ColoredString('')

            for elt in _in:
                color_in += strutil.ColoredString(elt[0], elt[1])

            rst = color_in.split(*args)

            dd('rst: ', rst)
            dd('expected: ', expect_rsts)
            self.assertEqual(rst, expect_rsts)

    def test_splitlines(self):
        color_cases = [
            ([('asd\r', 'red'), ('fer\n', 'blue'),  ('fg\r\n', 'white')],
             [True],
             [[('asd\r', 'red')], [('fer\n', 'blue')], [('fg\r\n', 'white')]],
             'test keepend true',
             ),

            ([('asd\r', 'red'), ('fer\n', 'blue'),  ('fg\r\n', 'white')],
             [False],
             [[('asd', 'red')], [('fer', 'blue')], [('fg', 'white')]],
             'test keepend false',
             ),

            ([('asd\r', 'red'), ('\nfer', 'blue'),  ('fg', 'white')],
             [True],
             [[('asd\r', 'red'), ('\n', 'blue')], [('fer', 'blue'), ('fg', 'white')]],
             '\\r\\n in diff color and keepend',
             ),

            ([('asd\r', 'red'), ('\nfer', 'blue'),  ('fg', 'white')],
             [False],
             [[('asd', 'red')], [('fer', 'blue'), ('fg', 'white')]],
             '\\r\\n in diff color and not keepend',
             ),

            ([('\nasd\r', 'red'), ('\nfer', 'blue'),  ('fg\r\n', 'white')],
             [True],
             [[('\n', 'red')], [('asd\r', 'red'), ('\n', 'blue')], [
                 ('fer', 'blue'), ('fg\r\n', 'white')]],
             'line break at the start and the end and keepend',
             ),

            ([('\nasd\r', 'red'), ('\nfer', 'blue'),  ('fg\r\n', 'white')],
             [False],
             [[('', 'red')], [('asd', 'red')], [
                 ('fer', 'blue'), ('fg', 'white')]],
             'line break at the start and the end and not keepend',
             ),

            ([('\n', 'red'), ('\r\n', 'blue'), ('\r\n', 'white')],
             [True],
             [[('\n', 'red')], [('\r\n', 'blue')], [('\r\n', 'white')]],
             'colored string consisted of all line breaks and keepend',
             ),

            ([('\n', 'red'), ('\r\n', 'blue'), ('\r\n', 'white')],
             [False],
             [[('', 'red')], [('', 'blue')], [('', 'white')]],
             'colored string consisted of all line breaks and not keepend',
             ),

            ([('asd ', 'red'), ('fer ', 'blue'),  ('fg', 'white')],
             [True],
             [[('asd ', 'red'), ('fer ', 'blue'),  ('fg', 'white')]],
             'no line break',
             ),

            ([('', 'red')],
             [True],
             [],
             'blank string',
             ),
        ]

        for _in, args, expected, msg in color_cases:
            dd('msg: ', msg)

            expect_rsts = []
            for elts in expected:
                cs = strutil.ColoredString('')
                for elt in elts:
                    cs += strutil.ColoredString(elt[0], elt[1])
                expect_rsts.append(cs)

            color_in = strutil.ColoredString('')
            for elt in _in:
                color_in += strutil.ColoredString(elt[0], elt[1])

            rst = color_in.splitlines(*args)

            dd('rst: ', rst)
            dd('expected: ', expect_rsts)
            self.assertEqual(rst, expect_rsts)

    def test_join(self):
        string_case = [
            ('ab',

             [('x', 'red'), ('y', 'blue')],

             [('a', None), ('x', 'red'), ('y', 'blue'), ('b', None)],

             'string iter and string element',
            ),

            ({'a':1, 'b':2},

             [('x', 'red'), ('y', 'blue')],

             [('a', None), ('x', 'red'), ('y', 'blue'), ('b', None)],

             'dict iter and string element',
            ),

            (['a', 'b', 'c'],

             [('x', 'red'), ('y', 'blue')],

             [('a', None), ('x', 'red'), ('y', 'blue'), ('b', None), ('x', 'red'), ('y', 'blue'), ('c', None)],

             'list iter and string element',
            ),

            (['a', 'b', 'c'],

             [('', None)],

             [('a', None), ('b', None), ('c', None)],

             'no separator',
            ),

            (['a'],

             [(' ', None)],

             [('a', None)],

             'iter with just 1 element',
            ),

            ([],

             [(' ', None)],

             [],

             'iter with no element',
            ),

        ]

        for iterable, sep, expected, msg in string_case:
            dd('msg: ', msg)

            color_sep = strutil.ColoredString('')
            for elt in sep:
                color_sep += strutil.ColoredString(elt[0], elt[1])

            rst = color_sep.join(iterable)

            color_expected = strutil.ColoredString('')
            for elt in expected:
                color_expected += strutil.ColoredString(elt[0], elt[1])

            dd('rst: ', rst)
            dd('expected: ', color_expected)
            self.assertEqual(rst, color_expected)


        ColoredString_case = [
            ([[('a', 'red'), ('b', 'blue')]],

             [('x', 'red'), ('y', 'blue')],

             [('a', 'red'), ('b', 'blue')],

             '1 ColoredString element',
            ),

            ([[('a', 'red')], [('b', 'blue'), ('c', 'white')]],

             [('x', 'red'), ('y', 'blue')],

             [('a', 'red'), ('x', 'red'), ('y', 'blue'), ('b', 'blue'), ('c', 'white')],

             '2 ColoredString elements',
             ),

            (['a', [('b', 'blue'), ('c', 'white')], 'd'],

             [('x', 'red'), ('y', 'blue')],

             [('a', None), ('x', 'red'), ('y', 'blue'), ('b', 'blue'), ('c', 'white'),
                 ('x', 'red'), ('y', 'blue'), ('d', None)],

             'ColoredString and string',
            ),

            ([[('a', 'red')], [('b', 'blue'), ('c', 'white')]],

             [(' ', None)],

             [('a', 'red'), (' ', None), ('b', 'blue'), ('c', 'white')],

             'with colored blank space',
            ),

            ([[('a', 'red')], [('b', 'blue')], [('c', 'white')]],

             [('', None)],

             [('a', 'red'), ('b', 'blue'), ('c', 'white')],

             'with no separator',
            ),

            ([[(' ', 'red')], [(' ', 'blue')]],

             [(' ', 'white')],

             [(' ', 'red'), (' ', 'white'),  (' ', 'blue')],

             'colored blank space element with blank space separator',
            ),

        ]

        for _in, sep, expected, msg in ColoredString_case:
            dd('msg: ', msg)

            color_in = []
            for l in _in:
                if type(l) != type([]):
                    color_in.append(l)
                    continue

                cs = strutil.ColoredString('')
                for elt in l:
                    cs += strutil.ColoredString(elt[0], elt[1])
                color_in.append(cs)

            color_sep = strutil.ColoredString('')
            for elt in sep:
                color_sep += strutil.ColoredString(elt[0], elt[1])

            rst = color_sep.join(color_in)

            color_expected = strutil.ColoredString('')
            for elt in expected:
                color_expected += strutil.ColoredString(elt[0], elt[1])

            dd('rst: ', rst)
            dd('expected: ', color_expected)
            self.assertEqual(rst, color_expected)



from .client import (
    EtcdException,
    EtcdInternalError,
    NoMoreMachineError,
    EtcdReadTimeoutError,
    EtcdRequestError,
    EtcdResponseError,
    EtcdIncompleteRead,
    EtcdSSLError,
    EtcdWatchError,
    EtcdKeyError,
    EtcdValueError,
    EcodeKeyNotFound,
    EcodeTestFailed,
    EcodeNotFile,
    EcodeNotDir,
    EcodeNodeExist,
    EcodeRootROnly,
    EcodeDirNotEmpty,
    EcodePrevValueRequired,
    EcodeTTLNaN,
    EcodeIndexNaN,
    EcodeInvalidField,
    EcodeInvalidForm,
    EcodeInscientPermissions,
    EtcdKeysResult,
    Response,
    EtcdError,
    Client,
)

__all__ = [
    "EtcdException",
    "EtcdInternalError",
    "NoMoreMachineError",
    "EtcdReadTimeoutError",
    "EtcdRequestError",
    "EtcdResponseError",
    "EtcdIncompleteRead",
    "EtcdSSLError",
    "EtcdWatchError",
    "EtcdKeyError",
    "EtcdValueError",
    "EcodeKeyNotFound",
    "EcodeTestFailed",
    "EcodeNotFile",
    "EcodeNotDir",
    "EcodeNodeExist",
    "EcodeRootROnly",
    "EcodeDirNotEmpty",
    "EcodePrevValueRequired",
    "EcodeTTLNaN",
    "EcodeIndexNaN",
    "EcodeInvalidField",
    "EcodeInvalidForm",
    "EcodeInscientPermissions",
    "EtcdKeysResult",
    "Response",
    "EtcdError",
    "Client",
]



import random
import unittest

from pykit import heap
from pykit import ututil

dd = ututil.dd


class X(object):

    def __init__(self, val):
        self.x = val

    def __lt__(self, b):
        return self.x < b.x

    def __str__(self):
        return str(self.x)


class TestRefHeap(unittest.TestCase):

    def test_index_level(self):
        cases = (
                (1, 0),
                (2, 1),
                (3, 1),
                (4, 2),
                (5, 2),
                (6, 2),
                (7, 2),
        )

        for inp, expected in cases:
            dd('inp:', inp)
            rst = heap.index_level(inp)

            dd('rst:', rst)
            self.assertEqual(expected, rst)

    def test_push_pop_get(self):

        cases = (
                ([], 0),
                ([None, 3, 1, None], 0),
                ([1, 1, 2], 0),
                (['a'], 0),
                (['a', 'c', 'b'], 0),
                ('accbbdb', 0),
                ([5, 3, 7, 6, 8, 1], 0),
                ([6, 4, 1, 7, 8, 3, 2], 0),
        )

        for inp, _ in cases:

            dd('inp:', inp)
            h = heap.RefHeap(inp)
            dd('after push')
            dd(str(h))

            self.assertEqual(len(inp), h.size)

            rst = []
            for i, _ in enumerate(inp):

                val = h.get()
                self.assertEqual(sorted(inp)[i], val)
                self.assertEqual(len(inp) - i, h.size)

                val = h.pop()
                self.assertEqual(sorted(inp)[i], val)
                self.assertEqual(len(inp) - i - 1, h.size)

                rst.append(val)

            dd('after pop all')
            dd(str(h))
            dd('rst:', rst)

            self.assertEqual(0, h.size)

    def test_class_with_lt(self):

        case = (6, 4, 1, 7, 8, 3, 2)
        arr = [X(x) for x in case]

        h = heap.RefHeap(arr)
        dd('after push:')
        dd(str(h))

        rst = h.pop_all(map=lambda x: x.x)
        self.assertEqual(sorted(case), rst)

    def test_sift_naive(self):

        a, b = [1], [2]
        h = heap.RefHeap([a, b])

        self.assertIs(a, h.get())

        a[0] = 3
        h.sift(a)
        self.assertIs(b, h.get())

        b[0] = 5
        h.sift(b)
        self.assertIs(a, h.get())

        self.assertRaises(heap.NotFound, h.sift, [])

    def test_sift(self):

        case = [6, 4, 1, 7, 8, 3, 2]

        for i in range(len(case)):
            for repl in (0, 5, 10):

                arr = [X(x) for x in case]
                dd('case:', case)

                h = heap.RefHeap(arr)
                dd('init heap:')
                dd(str(h))

                dd('replace {i}-th item {v} to {repl}'.format(i=i, v=case[i], repl=repl))

                arr[i].x = repl
                h.sift(arr[i])

                dd('after sift:')
                dd(str(h))

                rst = h.pop_all(map=lambda x: x.x)

                expected = case[:]
                expected[i] = repl
                expected.sort()

                self.assertEqual(expected, rst)

    def test_remove(self):

        case = [6, 4, 1, 7, 8, 3, 2]

        for i in range(len(case)):

            arr = [X(x) for x in case]
            dd('case:', case)

            h = heap.RefHeap(arr)
            dd('init heap:')
            dd(str(h))

            dd('remove {i}-th item {v}'.format(i=i, v=case[i]))

            h.remove(arr[i])

            dd('after remove:')
            dd(str(h))

            rst = h.pop_all(map=lambda x: x.x)

            expected = case[:i] + case[i + 1:]
            expected.sort()

            self.assertEqual(expected, rst)

    def test_remove_not_found(self):

        h = heap.RefHeap()
        self.assertRaises(heap.NotFound, h.remove, X(0))

        a = X(1)
        h.push(a)
        h.remove(a)

        self.assertRaises(heap.NotFound, h.remove, a)
        self.assertRaises(heap.NotFound, h.remove, X(0))

    def test_pop_all(self):

        case = [6, 4, 1, 7, 8, 3, 2]
        dd('case:', case)
        h = heap.RefHeap(case)
        self.assertEqual(sorted(case), h.pop_all())
        self.assertEqual(0, h.size)

        h = heap.RefHeap([X(x) for x in case])
        self.assertEqual(sorted(case), h.pop_all(map=lambda x: x.x))
        self.assertEqual(0, h.size)

    def test_dup(self):
        x = X(1)
        h = heap.RefHeap([x])

        self.assertRaises(heap.Duplicate, h.push, x)

        h.push(X(1))
        dd(str(h))

        self.assertEqual(1, h.pop().x)
        self.assertEqual(1, h.pop().x)

    def test_dup_all_type(self):

        for inp in [[1], {1: 2}, set([3])]:

            h = heap.RefHeap()
            h.push(inp)
            self.assertRaises(heap.Duplicate, h.push, inp)

    def test_pop_from_empty(self):
        h = heap.RefHeap()

        self.assertRaises(heap.Empty, h.pop)
        self.assertRaises(heap.Empty, h.get)

        h.push(1)
        h.pop()
        self.assertRaises(heap.Empty, h.pop)
        self.assertRaises(heap.Empty, h.get)

        self.assertIsNone(h.root)

    def test_rand(self):
        n = 500
        case = [random.randint(0, n) for x in range(n * 2)]
        h = heap.RefHeap(case)
        rst = []
        for x in range(len(case)):
            rst.append(h.pop())

        self.assertEqual(sorted(case), rst)



#!/usr/bin/env python2
# coding: utf-8

import os

import MySQLdb
import urllib

from pykit import dictutil
from pykit import mysqlconnpool


class ConnectionTypeError(Exception):
    pass


class InvalidLength(Exception):
    pass


def scan_index(connpool, table, result_fields, index_fields, index_values,
               left_open=False, limit=None, index_name=None, use_dict=True, retry=0):

    if type(connpool) == dict:
        # an address
        pool = mysqlconnpool.make(connpool)
    elif isinstance(connpool, mysqlconnpool.MysqlConnectionPool):
        pool = connpool
    else:
        raise ConnectionTypeError

    if len(index_values) != len(index_fields):
        raise InvalidLength(
            'number of index fields and values are not equal: '
            'index fields: {fld}; '
            'index values: {val}'.format(fld=index_fields, val=index_values))


    req_fields = list(index_fields)
    req_values = list(index_values)

    strict = True
    if limit is None:
        strict = False
        limit = 1024

    while True:
        sql = make_index_scan_sql(table, None, req_fields, req_values,
                             left_open=left_open, limit=limit, index_name=index_name)

        rst = pool.query(sql, retry=retry)

        for rr in rst:
            if use_dict:
                yield dict([(k, rr[k]) for k in result_fields])
            else:
                yield [rr[k] for k in result_fields]

        if strict:
            break

        if len(rst) > 0:
            last_row = rst[-1]
            req_fields = list(index_fields)
            req_values = [str(last_row[x]) for x in req_fields]
            left_open = True
            continue

        req_fields = req_fields[:-1]
        req_values = req_values[:-1]
        if len(req_fields) > 0:
            continue

        break


def make_index_scan_sql(table, result_fields, index, index_values, left_open=False, limit=1024, index_name=None):

    if left_open:
        operator = '>'
    else:
        operator = '>='

    if index_name is not None:
        force_index = index_name
    elif index is not None:
        force_index = 'idx_' + '_'.join(index)
    else:
        force_index = None

    return make_select_sql(table, result_fields, index, index_values, limit, force_index, operator)


def make_sql_condition(fld_vals, operator="=", formatter=list):

    cond_expressions = []

    for k, v in fld_vals[:-1]:
        cond_expressions.append(quote(k, "`") + " = " + _safe(v))

    key, value = fld_vals[-1]
    cond_expressions.append(quote(key, "`") + " " +
                            operator + " " + _safe(value))

    return formatter(cond_expressions)


def make_insert_sql(table, values, fields=None):

    sql_pattern = "INSERT INTO {tb}{fld_clause} VALUES {val_clause};"

    tb = make_table_name(table)

    if fields is not None:
        fld_clause = ' ({flds})'.format(
            flds=', '.join([quote(fld, "`") for fld in fields]))
    else:
        fld_clause = ''

    val_clause = '({vals})'.format(
        vals=', '.join([_safe(val) for val in values]))

    sql = sql_pattern.format(
        tb=tb, fld_clause=fld_clause, val_clause=val_clause)

    return sql


def make_update_sql(table, values, index, index_values, limit=None):

    sql_pattern = "UPDATE {tb} SET {set_clause}{where_clause}{limit_clause};"

    tb = make_table_name(table)

    set_clause = make_sql_condition(values.items(), formatter=', '.join)

    where_clause = make_where_clause(index, index_values)

    if limit is not None:
        limit_clause = ' LIMIT {n}'.format(n=limit)
    else:
        limit_clause = ''

    sql = sql_pattern.format(tb=tb, set_clause=set_clause, where_clause=where_clause,
                             limit_clause=limit_clause)

    return sql


def make_delete_sql(table, index, index_values, limit=None):

    sql_pattern = 'DELETE FROM {tb}{where_clause}{limit_clause};'

    tb = make_table_name(table)

    where_clause = make_where_clause(index, index_values)

    if limit is not None:
        limit_clause = ' LIMIT {n}'.format(n=limit)
    else:
        limit_clause = ''

    sql = sql_pattern.format(
        tb=tb, where_clause=where_clause, limit_clause=limit_clause)

    return sql


def make_select_sql(table, result_fields, index, index_values,
                    limit=None, force_index=None, operator='='):

    sql_pattern = 'SELECT {rst} FROM {tb}{force_index}{where_clause}{limit_clause};'

    tb = make_table_name(table)

    if result_fields is None:
        rst_flds = '*'
    else:
        rst_flds = ', '.join([quote(x, "`") for x in result_fields])

    if force_index is None:
        index_fld = ''
    else:
        index_fld = ' FORCE INDEX (`{idx}`)'.format(idx=force_index)

    where_clause = make_where_clause(index, index_values, operator)

    if limit is not None:
        limit_clause = ' LIMIT {n}'.format(n=limit)
    else:
        limit_clause = ''

    sql = sql_pattern.format(rst=rst_flds, tb=tb, force_index=index_fld, where_clause=where_clause,
                             limit_clause=limit_clause)

    return sql


def make_sql_range_conditions(fields, start, end=None):

    fld_len = len(fields)

    if len(start) != fld_len:
        raise InvalidLength(
                "the number of fields in 'start' and 'shard_fields' is not equal:"
                " start: {start}"
                " fields: {fields}".format(start=start, fields=fields))

    if end is None:
        end = type(start)([None] * len(start))

    elif len(end) != fld_len:
        raise InvalidLength(
                "the number of fields in 'end' and 'shard_fields' is not equal:"
                " end: {end}"
                " fields: {fields}".format(end=end, fields=fields))

    elif start >= end:
        return []

    fld_ranges = []
    for i in xrange(fld_len):
        fld_ranges.append((fields[i], (start[i], end[i]),))

    conditions = make_range_conditions(fld_ranges)

    result = []
    for cond in conditions:

        sql_conds = []
        for fld, operator, val in cond:
            sql_conds.append(quote(fld, '`') + operator + _safe(val))

        result.append(' AND '.join(sql_conds))

    return result


def make_range_conditions(fld_ranges, left_close=True):

    result = []
    start = 0
    end = 1

    # field[i] range is [start[i], end[i]), if start equals to end, it is a blank range.
    # continuose blank ranges in the beginning should not show as:
    # `fld > start[i] and fld < end[i]` but `fld = start[i]`.
    n_pref_blank_flds = 0
    for fld, _range in fld_ranges:
        if _range[start] == _range[end]:
            n_pref_blank_flds += 1
        else:
            break

    first_effective_fld = fld_ranges[n_pref_blank_flds][0]

    len_flds = len(fld_ranges)

    # init
    operator = ' > '
    range_use = start
    n_flds_use = len_flds

    while True:

        cond = []
        for fld, _range in fld_ranges[:n_flds_use - 1]:
            cond.append((fld, ' = ', _range[range_use]))

        fld, _range = fld_ranges[n_flds_use-1]

        if range_use == start and left_close:
            cond.append((fld, ' >= ', _range[start]))
            left_close = False
        else:
            cond.append((fld, operator, _range[range_use]))

        if fld == first_effective_fld:
            # no right boundary
            if _range[end] is None:
                result.append(cond)
                return result

            cond.append((fld, ' < ', _range[end]))

            operator = ' < '
            range_use = end

        if range_use == start:
            n_flds_use -= 1
        else:
            n_flds_use += 1

        result.append(cond)

        if n_flds_use > len_flds:
            break

    return result


def make_range_mysqldump_cmd(fields, conn, db, table, path_dump_to, dump_exec, start, end=None):

    cmd_pattern = ('{dump} --host={host} --port={port} --user={user} --password={passwd}'
                   ' {db} {table} -w {cond} > {rst_path}')

    if path_dump_to is None:
        rst_path = '{table}.sql'.format(table=urllib.quote_plus(table))
    elif isinstance(path_dump_to, basestring):
        rst_path = path_dump_to
    else:
        rst_path = os.path.join(*path_dump_to)

    if dump_exec is None:
        dump = 'mysqldump'
    elif isinstance(dump_exec, basestring):
        dump = dump_exec
    else:
        dump = os.path.join(*dump_exec)

    conditions = make_sql_range_conditions(fields, start, end)
    cond_expression = '(' + ') OR ('.join(conditions) + ')'

    pattern_args = dictutil.subdict(conn, ['host', 'port', 'user', 'passwd'],
                                    use_default=True, default='')

    pattern_args['dump'] = dump
    pattern_args['db'] = db
    pattern_args['table'] = table
    pattern_args['cond'] = cond_expression
    pattern_args['rst_path'] = rst_path

    for k, v in pattern_args.items():
        pattern_args[k] = quote(str(v), "'")

    return cmd_pattern.format(**pattern_args)


def quote(s, quote):

    if quote in s:
        s = s.replace(quote, "\\" + quote)

    return quote + s + quote


def _safe(s):
    return '"' + MySQLdb.escape_string(s) + '"'


def make_table_name(table_name):

    if isinstance(table_name, basestring):
        tb = quote(table_name, "`")
    else:
        db = quote(table_name[0], "`")
        tbl = quote(table_name[1], "`")
        tb = db + '.' + tbl

    return tb


def make_where_clause(index, index_values, operator='='):

    if index is not None:
        if len(index) != len(index_values):
            raise InvalidLength(
                'number of index fields and values are not equal: '
                'index fields: {fld}; '
                'index values: {val}'.format(fld=index, val=index_values))

        where_clause = ' WHERE {cond}'.format(
            cond=make_sql_condition(zip(index, index_values), operator, formatter=' AND '.join))
    else:
        where_clause = ''

    return where_clause



#!/usr/bin/env python2
# coding: utf-8

from collections import OrderedDict
from pykit.p3json.test import PyTest

from pykit import ututil

dd = ututil.dd


class TestS2Cases(object):

    def test_loads_dumps(self):

        cases = [
                ('""', ''),
                ('"it168/\xb6\xd4.exe"', 'it168/\xb6\xd4.exe'),
                ('"\xe6\x88\x91"', '我'),
        ]

        for json_str, obj in cases:

            mes = 'case json, object: ' + repr((json_str, obj))
            dd(mes)

            result = self.json.loads(json_str)
            self.assertEqual(obj, result)

            # XXX ensure_ascii=False: not to dump unicode as \uxxxx. Be compatible with lua cjson
            result = self.json.dumps(obj, ensure_ascii=False)
            self.assertEqual(json_str, result)

    def test_loads(self):

        cases = [
                ('"\\/\xb6\xd4."', '/\xb6\xd4.'),
                ('"\xe6\x88\x91"', '我'),

                # unicode is loaded and encoded in utf-8
                ('"abc\\u6211"', 'abc\xe6\x88\x91'),

                ('{"我": "我"}', {'我':'我'}),
                ('{"\\u6211": "\\u6211"}', {u'我'.encode('utf-8'): u'我'.encode('utf-8')}),
        ]

        for inp, expected in cases:

            mes = 'case load: json, object: ' + repr((inp, expected))
            dd(mes)

            result = self.json.loads(inp)
            dd('result:', repr(result))

            self.assertEqual(expected, result)

    def test_dumps(self):

        self.assertEqual('"\xe6\x88\x91"', self.json.dumps( '我', ensure_ascii=False))
        self.assertEqual('"\\u6211"',      self.json.dumps(u'我', ensure_ascii=True))
        self.assertEqual(u'"\u6211"',      self.json.dumps(u'我', ensure_ascii=False))
        self.assertEqual(str,              type(self.json.dumps(u'我', ensure_ascii=True)))
        self.assertEqual(str,              type(self.json.dumps( '我', ensure_ascii=False)))


class TestPyS2Cases(TestS2Cases, PyTest): pass



#!/usr/bin/env python
# coding: utf-8


class KeyValue(object):

    def create(self, key, value, ephemeral=False, sequence=False): raise TypeError('unimplemented')

    def delete(self, key, version=None): raise TypeError('unimplemented')

    def set(self, key, value, version=None): raise TypeError('unimplemented')

    def set_or_create(self, key, value, version=None): raise TypeError('unimplemented')

    def get(self, key): raise TypeError('unimplemented')


class Value(object):

    def create(self, value, ephemeral=False, sequence=False): raise TypeError('unimplemented')

    def delete(self, version=None): raise TypeError('unimplemented')

    def set(self, value, version=None): raise TypeError('unimplemented')

    def set_or_create(self, value, version=None): raise TypeError('unimplemented')

    def get(self): raise TypeError('unimplemented')




class IDBase(object):

    _tostr_fmt = '{item_1}-{item_2:0>3}'

    def __str__(self):
        return self._tostr_fmt.format(**{k: str(v) for k, v in self._asdict().items()})

    def tostr(self):
        return str(self)



#!/usr/bin/env python2
# coding: utf-8

import os
import random
import time
import unittest

import websocket

from pykit import proc
from pykit import utfjson
from pykit import ututil
from pykit.wsjobd import Job
from pykit.wsjobd.test.wsjobd_server import PORT

dd = ututil.dd

random.seed(time.time())
this_base = os.path.dirname(__file__)


def subproc(script):
    return proc.command('sh', close_fds=True,
                        stdin=script)


class TestWsjobd(unittest.TestCase):

    @classmethod
    def _clean(cls):
        try:
            subproc('python2 {b}/wsjobd_server.py stop'.format(b=this_base))
        except Exception as e:
            dd('failed to stop wsjobd server: ' + repr(e))

        time.sleep(0.1)

    @classmethod
    def setUpClass(cls):
        cls._clean()
        subproc('python2 {b}/wsjobd_server.py start'.format(b=this_base))
        time.sleep(1)

    @classmethod
    def tearDownClass(cls):
        cls._clean()

    def _create_client(self):
        ws = websocket.WebSocket()
        ws.connect('ws://127.0.0.1:%d' % PORT)
        ws.timeout = 6
        return ws

    def setUp(self):
        self.ws = self._create_client()

    def tearDown(self):
        self.ws.close()

    def get_random_ident(self):
        return 'random_ident_%d' % random.randint(10000, 99999)

    def test_invalid_jobdesc(self):
        cases = (
            ('foo', 'not json'),
            (utfjson.dump('foo'), 'not dict'),
            (utfjson.dump({}), 'no func'),
            (utfjson.dump({'func': 'foo'}), 'no ident'),
            (utfjson.dump({'ident': 'bar'}), 'no func'),
            (utfjson.dump({'ident': 'bar', 'func': {}}), 'invalid func'),
            (utfjson.dump({'ident': 44, 'func': 'foo'}), 'invalid ident'),
            (utfjson.dump({'ident': 'foo', 'func': 'foo', 'jobs_dir': {}}),
             'invalid jobs_dir'),
        )
        for msg, desc in cases:
            ws = self._create_client()
            ws.send(msg)
            resp = utfjson.load(ws.recv())
            self.assertIn('err', resp, desc)
            ws.close()

    def test_normal_job(self):
        job_desc = {
            'func': 'test_job_normal.run',
            'ident': self.get_random_ident(),
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        resp = utfjson.load(self.ws.recv())
        self.assertEqual('foo', resp['result'], 'test get result')

    def test_report_interval(self):
        job_desc = {
            'func': 'test_job_loop_10.run',
            'ident': self.get_random_ident(),
            'progress': {
                'interval': 0.5,
            },
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        self.ws.recv()
        first_report_time = time.time()

        self.ws.recv()
        second_report_time = time.time()

        actual_interval = second_report_time - first_report_time

        diff = actual_interval - job_desc['progress']['interval']

        # tolerate 0.1 second of difference
        self.assertLess(diff, 0.1)

    def test_progress_key(self):
        job_desc = {
            'func': 'test_job_progress_key.run',
            'ident': self.get_random_ident(),
            'progress': {
                'key': 'foo',
            },
            'report_system_load': True,
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        resp = self.ws.recv()
        resp = utfjson.load(resp)
        self.assertEqual('80%', resp)

    def test_check_system_load(self):
        job_desc = {
            'func': 'test_job_normal.run',
            'ident': self.get_random_ident(),
            'check_load': {
                'mem_low_threshold': 100 * 1024 ** 3,
                'cpu_low_threshold': 0,
            },
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        resp = self.ws.recv()
        resp = utfjson.load(resp)
        self.assertEqual('SystemOverloadError', resp['err'])

        job_desc = {
            'func': 'test_job_normal.run',
            'ident': self.get_random_ident(),
            'check_load': {
                'cpu_low_threshold': 100.1,
                'mem_low_threshold': 0,
            },
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        ws2 = self._create_client()
        ws2.send(utfjson.dump(job_desc))

        resp = ws2.recv()
        resp = utfjson.load(resp)
        ws2.close()
        self.assertEqual('SystemOverloadError', resp['err'])

    def test_max_client_number(self):
        job_desc = {
            'func': 'test_job_loop_10.run',
            'ident': self.get_random_ident(),
            'check_load': {
                'max_client_number': 1,
                'cpu_low_threshold': 0,
                'mem_low_threshold': 0,
            },
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        resp = self.ws.recv()
        resp = utfjson.load(resp)
        self.assertNotIn('err', resp)

        job_desc = {
            'func': 'test_job_loop_10.run',
            'ident': self.get_random_ident(),
            'check_load': {
                'max_client_number': 1,
                'cpu_low_threshold': 0,
                'mem_low_threshold': 0,
            },
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        ws2 = self._create_client()
        ws2.send(utfjson.dump(job_desc))

        resp = ws2.recv()
        resp = utfjson.load(resp)
        self.assertEqual('SystemOverloadError', resp['err'])
        ws2.close()

    def test_module_not_exists(self):
        job_desc = {
            'func': 'foo.bar',
            'ident': self.get_random_ident(),
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        resp = utfjson.load(self.ws.recv())
        self.assertIn('err', resp)

    def test_report_system_load(self):
        job_desc = {
            'func': 'test_job_normal.run',
            'ident': self.get_random_ident(),
            'report_system_load': True,
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        resp = utfjson.load(self.ws.recv())
        self.assertIn('mem_available', resp['system_load'])
        self.assertIn('cpu_idle_percent', resp['system_load'])
        self.assertIn('client_number', resp['system_load'])

    def test_same_ident_same_job(self):
        ident = self.get_random_ident()
        job_desc = {
            'func': 'test_job_echo.run',
            'ident': ident,
            'echo': 'foo',
            'sleep_time': 10,
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        resp = utfjson.load(self.ws.recv())
        self.assertEqual('foo', resp['result'])

        job_desc = {
            'func': 'test_job_echo.run',
            'ident': ident,
            'echo': 'bar',
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        ws = self._create_client()
        ws.send(utfjson.dump(job_desc))

        resp = utfjson.load(ws.recv())
        ws.close()
        # not bar, because the ident is same as the first job,
        # if job exists, it will not create a new one
        self.assertEqual('foo', resp['result'])
        self.assertEqual('foo', resp['echo'])

    def test_same_ident_different_job(self):
        ident = self.get_random_ident()
        job_desc = {
            'func': 'test_job_echo.run',
            'ident': ident,
            'echo': 'foo',
            'sleep_time': 0.1,
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        resp = utfjson.load(self.ws.recv())
        self.assertEqual('foo', resp['result'])

        time.sleep(0.2)

        job_desc = {
            'func': 'test_job_echo.run',
            'ident': ident,
            'echo': 'bar',
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        ws2 = self._create_client()
        ws2.send(utfjson.dump(job_desc))

        resp = utfjson.load(ws2.recv())
        ws2.close()
        # old job with the same ident has exit, it will create a new one
        self.assertEqual('bar', resp['result'])
        self.assertEqual('bar', resp['echo'])

    def test_client_close(self):
        ident = self.get_random_ident()
        job_desc = {
            'func': 'test_job_echo.run',
            'ident': ident,
            'echo': 'foo',
            'time_sleep': 10,
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))
        self.ws.close()
        self.ws = self._create_client()

        job_desc = {
            'func': 'test_job_echo.run',
            'ident': ident,
            'echo': 'bar',
            'time_sleep': 10,
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        resp = self.ws.recv()
        resp = utfjson.load(resp)
        self.assertEqual('foo', resp['result'])

    def test_invalid_cpu_sample_interval(self):
        job_desc = {
            'func': 'test_job_normal.run',
            'ident': self.get_random_ident(),
            'cpu_sample_interval': 'foo',
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))

        resp = utfjson.load(self.ws.recv())
        self.assertEqual('InvalidMessageError', resp['err'])

    def test_invalid_check_load_args(self):
        cases = (
            {'check_load': {'mem_low_threshold': 'foo',
                            'cpu_low_threshold': 0}},
            {'check_load': {'cpu_low_threshold': None,
                            'mem_low_threshold': 0}},
            {'check_load': {'max_client_number': {},
                            'cpu_low_threshold': 0,
                            'mem_low_threshold': 0}},
        )

        job_desc = {
            'func': 'test_job_normal.run',
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        for case in cases:
            case.update(job_desc)
            case['ident'] = self.get_random_ident()

            ws = self._create_client()
            ws.send(utfjson.dump(case))

            resp = utfjson.load(ws.recv())
            ws.close()
            self.assertEqual('InvalidMessageError', resp['err'])

    def test_func_not_exists(self):
        ident = self.get_random_ident()
        job_desc = {
            'func': 'test_job_echo.func_not_exists',
            'ident': ident,
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
        }

        self.ws.send(utfjson.dump(job_desc))
        resp = self.ws.recv()
        resp = utfjson.load(resp)
        self.assertEqual('LoadingError', resp['err'])

    def test_worker_exception(self):
        ident = self.get_random_ident()
        job_desc = {
            'func': 'test_job_worker_exception.run',
            'ident': ident,
            'jobs_dir': 'pykit/wsjobd/test/test_jobs',
            'progress': {
                'interval': 0.1,
            },
        }

        self.ws.send(utfjson.dump(job_desc))

        for i in range(10):
            resp = self.ws.recv()
            self.assertNotIn('err', resp)

        with self.assertRaises(Exception):
            for i in range(3):
                self.ws.recv()

    def test_create_job(self):
        self.assertEqual(0, len(Job.sessions))

        def f(self):
            time.sleep(0.2)
            return

        Job('channel', {'ident': 'a'}, f)
        joba = Job.sessions['a']

        self.assertEqual(1, len(Job.sessions))

        Job('channel', {'ident': 'a'}, f)
        joba1 = Job.sessions['a']

        #  joba already exists
        self.assertEqual(joba1, joba)
        self.assertEqual(1, len(Job.sessions))

        time.sleep(0.15)
        Job('channel', {'ident': 'b'}, f)

        self.assertEqual(2, len(Job.sessions))

        time.sleep(0.15)
        self.assertEqual(1, len(Job.sessions))

        time.sleep(0.1)
        self.assertEqual(0, len(Job.sessions))

        #  test use same ident after first one exit
        Job('channel', {'ident': 'a'}, f)
        joba1 = Job.sessions['a']

        time.sleep(0.25)

        Job('channel', {'ident': 'a'}, f)
        joba2 = Job.sessions['a']
        self.assertNotEqual(joba1, joba2)



#!/usr/bin/env python2
# coding: utf-8

import time
import unittest

import docker

from pykit import mysqlconnpool
from pykit import mysqlutil
from pykit import proc
from pykit import ututil

dd = ututil.dd

mysql_test_password = '123qwe'
mysql_test_port = 3306
mysql_test_user = 'root'
mysql_test_name = 'mysql_test'
mysql_test_tag = 'test-mysql:0.0.1'
mysql_test_db = 'test'
mysql_test_table = 'errlog'


class TestMysqlutil(unittest.TestCase):

    def test_scan_index(self):

        mysql_ip = start_mysql_server()

        addr = {
            'host': mysql_ip,
            'port': mysql_test_port,
            'user': mysql_test_user,
            'passwd': mysql_test_password,
        }

        conns = (addr,
                 mysqlconnpool.make(addr),
                )

        table = ('test', 'errlog')

        result_fields = ['_id']

        cases = (
            ([['service', 'ip', '_id'], ['common0', '127.0.0.1', '8']],
             {},
             ('8', '12', '18', '20', '32', '2', '3', '13', '19', '27', '30', '11', '22', '29', '31', '14',
              '28', '4', '5', '9', '24', '6', '15', '21', '23', '25', '26', '10', '16', '7', '17'),

             'test common',
            ),
            ([['service', 'ip', '_id'], ['common0', '127.0.0.1', '8']],
             {'left_open': True},
             ('12', '18', '20', '32', '2', '3', '13', '19', '27', '30', '11', '22', '29', '31', '14',
              '28', '4', '5', '9', '24', '6', '15', '21', '23', '25', '26', '10', '16', '7', '17'),

             'test left_open',
            ),
            ([['service', 'ip', '_id'], ['common0', '127.0.0.1', '8']],
             {'limit': 3},
             ('8', '12', '18', ),

             'test limit',
            ),
            ([['autolvl', 'service', 'ip', '_id'], ['stable', 'common0', '127.0.0.1', '8']],
             {'index_name': 'idx_service_ip__id'},
             ('12', '32', '2', '13', '19', '30', '22', '28', '6', '15', '21', '7'),

             'test index_name',
            ),
            ([['service', 'ip', '_id'], ['common0', '127.0.0.1', '8']],
             {'left_open': True, 'limit': 3, 'index_name': 'idx_time__id'},
             ('12', '18', '20', ),

             'test all kwargs',
            ),
        )

        try:
            for conn in conns:
                dd('conn: ', conn)

                for args, kwargs, rst_expect, msg in cases:

                    args = [conn, table, result_fields] + args
                    kwargs['use_dict'] = False

                    dd('msg: ', msg)

                    rst = mysqlutil.scan_index(*args, **kwargs)

                    for i, rr in enumerate(rst):
                        dd('rst:', rr)
                        dd('except: ', rst_expect[i])

                        self.assertEqual(rr[0], long(rst_expect[i]))

                    self.assertEqual(len(rst_expect), i+1)
        finally:
            stop_mysql_server()

        error_cases = (
            ([addr, table, result_fields, ['service', 'ip', '_id'], ['common0', '127.0.0.2']],
             {},
             mysqlutil.InvalidLength,

             'test index_fields amount greater than index_values',
             ),
            ([addr, table, result_fields, ['service', 'ip'], ['common0', '127.0.0.2', '13']],
             {},
             mysqlutil.InvalidLength,

             'test index_fields amount less than index_values',
             ),
            (['addr', table, result_fields, ['service', 'ip', '_id'], ['common0', '127.0.0.1', '8']],
             {},
             mysqlutil.ConnectionTypeError,

             'test conn type error',
             ),
        )

        for args, kwargs, error, msg in error_cases:
            dd('msg: ', msg)

            try:
                rst = mysqlutil.scan_index(*args, **kwargs)
            except error as e:
                self.assertEqual(type(e), error)

    def test_make_index_scan_sql(self):

        table = 'errlog'

        cases = (
            ([['_id'], ['service', 'ip', '_id'], ['common0', '127.0.0.1', '8']],
             {},
             'SELECT `_id` FROM `errlog` FORCE INDEX (`idx_service_ip__id`) '
             'WHERE `service` = "common0" AND `ip` = "127.0.0.1" AND `_id` >= "8" LIMIT 1024;',

             'test common',
            ),
            ([['_id', 'ip'], ['service', 'ip', '_id'], ['common0', '127.0.0.1', '8']],
             {'left_open': True},
             'SELECT `_id`, `ip` FROM `errlog` FORCE INDEX (`idx_service_ip__id`) '
             'WHERE `service` = "common0" AND `ip` = "127.0.0.1" AND `_id` > "8" LIMIT 1024;',

             'test left_open',
            ),
            ([None, ['service', 'ip', '_id'], ['common0', '127.0.0.1', '8']],
             {'limit': 3},
             'SELECT * FROM `errlog` FORCE INDEX (`idx_service_ip__id`) '
             'WHERE `service` = "common0" AND `ip` = "127.0.0.1" AND `_id` >= "8" LIMIT 3;',

             'test limit',
            ),
            ([['_id'], ['autolvl', 'service', 'ip', '_id'], ['stable', 'common0', '127.0.0.1', '8']],
             {'index_name': 'idx_service_ip__id'},
             'SELECT `_id` FROM `errlog` FORCE INDEX (`idx_service_ip__id`) '
             'WHERE `autolvl` = "stable" AND `service` = "common0" AND `ip` = "127.0.0.1" '
             'AND `_id` >= "8" LIMIT 1024;',

             'test index_name',
            ),
            ([['_id'], None, ['stable', 'common0', '127.0.0.1', '8']],
             {},
             'SELECT `_id` FROM `errlog` LIMIT 1024;',

             'test blank index_fields',
            ),
            ([['_id'], ['service', 'ip', '_id'], ['common0', '127.0.0.1', '8']],
             {'left_open': True, 'limit': 5, 'index_name': 'idx_time__id'},
             'SELECT `_id` FROM `errlog` FORCE INDEX (`idx_time__id`) '
             'WHERE `service` = "common0" AND `ip` = "127.0.0.1" AND `_id` > "8" LIMIT 5;',

             'test all kwargs',
            ),
        )

        for args, kwargs, rst_expect, msg in cases:

            dd('msg: ', msg)
            dd('rst_expect: ', rst_expect)

            args = [table] + args

            rst = mysqlutil.make_index_scan_sql(*args, **kwargs)

            dd('rst       : ', rst)

            self.assertEquals(rst, rst_expect)

    def test_make_insert_sql(self):

        cases = (
            (
                mysql_test_table,
                ['common1', '127.0.0.3', '3'],
                None,

                'INSERT INTO `errlog` VALUES ("common1", "127.0.0.3", "3");',
            ),

            (
                mysql_test_table,
                ['common0', '127.0.0.4', '2'],
                ['service', 'ip', '_id'],

                'INSERT INTO `errlog` (`service`, `ip`, `_id`) '
                'VALUES ("common0", "127.0.0.4", "2");',
            ),

            (
                (mysql_test_db, mysql_test_table),
                ['common2', '127.0.0.3', '4'],
                None,

                'INSERT INTO `test`.`errlog` VALUES ("common2", "127.0.0.3", "4");',
            ),

            (
                (mysql_test_db, mysql_test_table),
                ['common"三"', '127.0.0.3', '\\"4'],
                None,

                'INSERT INTO `test`.`errlog` VALUES ("common\\\"三\\\"", "127.0.0.3", "\\\\\\"4");',
            ),
        )

        for table, values, fields, expected in cases:

            dd('expected: ', expected)
            rst = mysqlutil.make_insert_sql(table, values, fields)
            dd('result:   ', rst)

            self.assertEqual(expected, rst)

    def test_make_select_sql(self):

        cases = (
            (
                mysql_test_table,
                ['_id'],
                ['service', 'ip', '_id'],
                ['common0', '127.0.0.1', '8'],
                {},
                'SELECT `_id` FROM `errlog` WHERE `service` = "common0" AND `ip` = "127.0.0.1" '
                'AND `_id` = "8";',
            ),

            (
                (mysql_test_db, mysql_test_table),
                ['_id', 'ip'],
                ['service', 'ip', '_id'],
                ['common0', '127.0.0.1', '8'],
                {
                    'limit': 1024,
                },

                'SELECT `_id`, `ip` FROM `test`.`errlog` WHERE `service` = "common0" '
                'AND `ip` = "127.0.0.1" AND `_id` = "8" LIMIT 1024;',
            ),

            (
                mysql_test_table,
                None,
                ['service', 'ip', '_id'],
                ['common0', '127.0.0.1', '8'],
                {
                    'limit': 3,
                    'force_index': 'idx_service_ip__id',
                },

                'SELECT * FROM `errlog` FORCE INDEX (`idx_service_ip__id`) '
                'WHERE `service` = "common0" AND `ip` = "127.0.0.1" AND `_id` = "8" LIMIT 3;',
            ),

            (
                mysql_test_table,
                ['_id'],
                ['service', 'ip', '_id'],
                ['common0', '127.0.0.1', '8'],
                {
                    'limit': 3,
                    'force_index': 'idx_service_ip',
                    'operator': '>=',
                },

                'SELECT `_id` FROM `errlog` FORCE INDEX (`idx_service_ip`) '
                'WHERE `service` = "common0" AND `ip` = "127.0.0.1" AND `_id` >= "8" LIMIT 3;',
            ),

            (
                mysql_test_table,
                ['_id'],
                None,
                None,
                {},

                'SELECT `_id` FROM `errlog`;',
            ),

            (
                mysql_test_table,
                ['_id'],
                ['service', 'ip', '_id'],
                ['common"三"', '127.0.0.1', '\\\'8'],
                {
                    'limit': 3,
                    'force_index': 'idx_service_ip',
                    'operator': '>=',
                },

                'SELECT `_id` FROM `errlog` FORCE INDEX (`idx_service_ip`) '
                'WHERE `service` = "common\\\"三\\\"" AND `ip` = "127.0.0.1" '
                'AND `_id` >= "\\\\\\\'8" LIMIT 3;',
            ),
        )

        for table, result_fields, index, index_values, kwargs, expected in cases:

            dd('expected: ', expected)
            rst = mysqlutil.make_select_sql(
                table, result_fields, index, index_values, **kwargs)
            dd('result:   ', rst)

            self.assertEqual(expected, rst)

    def test_make_update_sql(self):

        cases = (
            (
                mysql_test_table,
                {
                    '_id': '0',
                    'time': '042718',
                },
                ['service', 'ip', '_id'],
                ['common0', '127.0.0.1', '8'],
                {},
                'UPDATE `errlog` SET `_id` = "0", `time` = "042718" '
                'WHERE `service` = "common0" AND `ip` = "127.0.0.1" AND `_id` = "8";',
            ),

            (
                (mysql_test_db, mysql_test_table),
                {
                    '_id': '0',
                    'time': '042718',
                },
                ['service', 'ip', '_id'],
                ['common0', '127.0.0.1', '8'],
                {
                    'limit': 1,
                },
                'UPDATE `test`.`errlog` SET `_id` = "0", `time` = "042718" '
                'WHERE `service` = "common0" AND `ip` = "127.0.0.1" AND `_id` = "8" LIMIT 1;',
            ),

            (
                (mysql_test_db, mysql_test_table),
                {
                    '_id': '0',
                    'time': '042718',
                },
                None,
                None,
                {
                    'limit': 1,
                },
                'UPDATE `test`.`errlog` SET `_id` = "0", `time` = "042718" LIMIT 1;',
            ),

            (
                (mysql_test_db, mysql_test_table),
                {
                    '_id': '0',
                    'time': '04\"27\'18',
                },
                ['service', 'ip', '_id'],
                ['common"三"', '127.0.0.1', '8'],
                {
                    'limit': 1,
                },
                'UPDATE `test`.`errlog` SET `_id` = "0", `time` = "04\\\"27\\\'18" '
                'WHERE `service` = "common\\\"三\\\"" AND `ip` = "127.0.0.1" AND `_id` = "8" LIMIT 1;',
            ),
        )

        for table, values, index, index_values, kwargs, expected in cases:

            dd('expected: ', expected)
            rst = mysqlutil.make_update_sql(
                table, values, index, index_values, **kwargs)
            dd('result:   ', rst)

            self.assertEqual(expected, rst)

    def test_make_delete_sql(self):

        cases = (
            (
                mysql_test_table,
                ['service', 'ip', '_id'],
                ['common0', '127.0.0.1', '8'],
                {},
                'DELETE FROM `errlog` WHERE `service` = "common0" AND `ip` = "127.0.0.1" '
                'AND `_id` = "8";',
            ),

            (
                (mysql_test_db, mysql_test_table),
                ['service', 'ip', '_id'],
                ['common0', '127.0.0.1', '8'],
                {
                    'limit': 1,
                },

                'DELETE FROM `test`.`errlog` WHERE `service` = "common0" '
                'AND `ip` = "127.0.0.1" AND `_id` = "8" LIMIT 1;',
            ),

            (
                mysql_test_table,
                None,
                None,
                {
                    'limit': 1,
                },

                'DELETE FROM `errlog` LIMIT 1;',
            ),

            (
                mysql_test_table,
                ['service', 'ip', '_id'],
                ['common"三"', '127.0.0.1', '\\\'8'],
                {
                    'limit': 1,
                },

                'DELETE FROM `errlog` '
                'WHERE `service` = "common\\\"三\\\"" AND `ip` = "127.0.0.1" '
                'AND `_id` = "\\\\\\\'8" LIMIT 1;',
            ),
        )

        for table, index, index_values, kwargs, expected in cases:

            dd('expected: ', expected)
            rst = mysqlutil.make_delete_sql(
                table, index, index_values, **kwargs)
            dd('result:   ', rst)

            self.assertEqual(expected, rst)

    def test_make_sql_range_conditions(self):

        cases = (
            ((['id'], ['10']),
                ['`id` >= "10"'],
                '1 shard field, no end shard'),

            ((['id'], ['10'], ['15']),
                ['`id` >= "10" AND `id` < "15"'],
                '1 shard field normal'),

            ((['id', 'service'], ['10', 'a']),
                ['`id` = "10" AND `service` >= "a"',
                 '`id` > "10"',
                 ],
                '2 shard fields, no end shard'),

            ((['id', 'service'], ['10', 'a'], ['10', 'd']),
                ['`id` = "10" AND `service` >= "a" AND `service` < "d"'],
                '2 shard field, 1 same fields'),

            ((['id', 'service'], ['10', 'a'], ['15', 'd']),
                ['`id` = "10" AND `service` >= "a"',
                 '`id` > "10" AND `id` < "15"',
                 '`id` = "15" AND `service` < "d"',
                 ],
                '2 shard fields normal'),

            ((['id', 'service', 'level'], ['10', 'a', 'a3']),
                ['`id` = "10" AND `service` = "a" AND `level` >= "a3"',
                 '`id` = "10" AND `service` > "a"',
                 '`id` > "10"',
                 ],
                '3 shard fields, no end shard'),

            ((['id', 'service', 'level'], ['10', 'a', 'a3'], ['10', 'a', 'b2']),
                ['`id` = "10" AND `service` = "a" AND `level` >= "a3" AND `level` < "b2"'],
                '3 shard fields, 2 same fields'),

            ((['id', 'service', 'level'], ['10', 'a', 'a3'], ['10', 'd', 'b2']),
                ['`id` = "10" AND `service` = "a" AND `level` >= "a3"',
                 '`id` = "10" AND `service` > "a" AND `service` < "d"',
                 '`id` = "10" AND `service` = "d" AND `level` < "b2"',
                 ],
                '3 shard fields, 1 same fields'),

            ((['id', 'service', 'level'], ['10', 'a', 'a3'], ['15', 'd', 'b2']),
                ['`id` = "10" AND `service` = "a" AND `level` >= "a3"',
                 '`id` = "10" AND `service` > "a"',
                 '`id` > "10" AND `id` < "15"',
                 '`id` = "15" AND `service` < "d"',
                 '`id` = "15" AND `service` = "d" AND `level` < "b2"',
                 ],
                '3 shard fields normal'),

            ((['id', 'service', 'level'], ['15', 'd', 'b2'], ['10', 'a', 'a3']),
                [],
                'start > end'),

            ((['id', 'service', 'level'], ['10', 'a', 'a3'], ['10', 'a', 'a3']),
                [],
                'start == end'),

            ((['id', 'service', 'level'], ['10', 'a"c', 'a"3"'], ['15', 'd\\b', 'b\'2']),
                ['`id` = "10" AND `service` = "a\\\"c" AND `level` >= "a\\\"3\\\""',
                 '`id` = "10" AND `service` > "a\\\"c"',
                 '`id` > "10" AND `id` < "15"',
                 '`id` = "15" AND `service` < "d\\\\b"',
                 '`id` = "15" AND `service` = "d\\\\b" AND `level` < "b\\\'2"',
                 ],
                'special characters'),
        )

        for args, rst_expected, msg in cases:
            dd('msg: ', msg)
            dd('expected: ', rst_expected)

            rst = mysqlutil.make_sql_range_conditions(*args)
            dd('rst     : ', rst)

            self.assertEqual(rst, rst_expected)

    def test_make_range_mysqldump_cmd(self):

        conn = {
            'host': '127.0.0.1',
            'user': 'root',
            'passwd': 'password',
            'port': 3306,
        }

        conn_special = {
            'host': '127.0.0.1',
            'user': 'root"1',
            'passwd': 'pass\'word',
            'port': 3306,
        }

        db = 'mysql'
        table = 'key'
        table_special = 'key"00"'

        cases = (
            ((['id', 'service'], conn, db, table, ['/tmp', 'key.sql'], ['/usr', 'bin', 'mysqldump'],
                ['10', 'a'], ['15', 'd']),
                ("'/usr/bin/mysqldump' --host='127.0.0.1' --port='3306' --user='root' --password='password' " +
                 "'mysql' 'key' -w '" +
                 "(`id` = \"10\" AND `service` >= \"a\") OR " +
                 "(`id` > \"10\" AND `id` < \"15\") OR " +
                 "(`id` = \"15\" AND `service` < \"d\")" +
                 "' > '/tmp/key.sql'"),
                'test normal'
             ),

            ((['id', 'service'], conn, db, table, '/tmp/key.sql', ['/usr', 'bin', 'mysqldump'],
                ['10', 'a'], ['15', 'd']),
                ("'/usr/bin/mysqldump' --host='127.0.0.1' --port='3306' --user='root' --password='password' " +
                 "'mysql' 'key' -w '" +
                 "(`id` = \"10\" AND `service` >= \"a\") OR " +
                 "(`id` > \"10\" AND `id` < \"15\") OR " +
                 "(`id` = \"15\" AND `service` < \"d\")" +
                 "' > '/tmp/key.sql'"),
                'test string path_dump_to'
             ),

            ((['id', 'service'], conn, db, table, ['/tmp', 'key.sql'], '/usr/bin/mysqldump',
                ['10', 'a'], ['15', 'd']),
                ("'/usr/bin/mysqldump' --host='127.0.0.1' --port='3306' --user='root' --password='password' " +
                 "'mysql' 'key' -w '" +
                 "(`id` = \"10\" AND `service` >= \"a\") OR " +
                 "(`id` > \"10\" AND `id` < \"15\") OR " +
                 "(`id` = \"15\" AND `service` < \"d\")" +
                 "' > '/tmp/key.sql'"),
                'test string mysqldump path'
             ),

            ((['id', 'service'], conn, db, table, ['/tmp', 'key.sql'], ['/usr', 'bin', 'mysqldump'],
                ['10', 'a']),
                ("'/usr/bin/mysqldump' --host='127.0.0.1' --port='3306' --user='root' --password='password' " +
                 "'mysql' 'key' -w '" +
                 "(`id` = \"10\" AND `service` >= \"a\") OR " +
                 "(`id` > \"10\")" +
                 "' > '/tmp/key.sql'"),
                'test no end shard'
             ),

            ((['id', 'service'], conn, db, table, ['/tmp', 'key.sql'], ['/usr', 'bin', 'mysqldump'],
                ['10', 'a"'], ['15', 'd"3"']),
                ("'/usr/bin/mysqldump' --host='127.0.0.1' --port='3306' --user='root' --password='password' " +
                 "'mysql' 'key' -w '"
                 "(`id` = \"10\" AND `service` >= \"a\\\"\") OR " +
                 "(`id` > \"10\" AND `id` < \"15\") OR " +
                 "(`id` = \"15\" AND `service` < \"d\\\"3\\\"\")" +
                 "' > '/tmp/key.sql'"),
                'test special characters in shards'
             ),

            ((['id', 'service'], conn_special, db, table_special, ['/tmp', 'key.sql'], ['/usr', 'bin', 'mysqldump'],
                ['10', 'a'], ['15', 'd']),
                ("'/usr/bin/mysqldump' --host='127.0.0.1' --port='3306' --user='root\"1' --password='pass\\\'word' " +
                 "'mysql' 'key\"00\"' -w '" +
                 "(`id` = \"10\" AND `service` >= \"a\") OR " +
                 "(`id` > \"10\" AND `id` < \"15\") OR " +
                 "(`id` = \"15\" AND `service` < \"d\")" +
                 "' > '/tmp/key.sql'"),
                'test special characters in dbinfo'
             ),
        )

        for args, rst_expected, msg in cases:
            dd('msg: ', msg)
            dd('expected: ', rst_expected)
            rst = mysqlutil.make_range_mysqldump_cmd(*args)
            dd('rst     : ', rst)

            self.assertEquals(rst_expected, rst)

def docker_does_container_exist(name):

    dcli = _docker_cli()
    try:
        dcli.inspect_container(name)
        return True
    except docker.errors.NotFound:
        return False

def _docker_cli():
    dcli = docker.Client(base_url='unix://var/run/docker.sock')
    return dcli

def start_mysql_server():

    # create docker image by run mysqlutil/test/dep/build_img.sh before test
    if not docker_does_container_exist(mysql_test_name):

        dd('create container: ' + mysql_test_name)
        dcli = _docker_cli()
        dcli.create_container(name=mysql_test_name,
                              environment={
                                  'MYSQL_ROOT_PASSWORD': mysql_test_password,
                              },
                              image=mysql_test_tag,
                              )
        time.sleep(2)

    dd('start mysql: ' + mysql_test_name)
    dcli = _docker_cli()
    dcli.start(container=mysql_test_name)

    dd('get mysql ip inside container')
    rc, out, err = proc.command(
        'docker',
        'run',
        '-i',
        '--link', mysql_test_name + ':mysql',
        '--rm', mysql_test_tag,
        'sh', '-c', 'exec echo "$MYSQL_PORT_3306_TCP_ADDR"',
    )

    ip = out.strip()
    dd('ip: ' + repr(ip))

    return ip

def stop_mysql_server():

    # remove docker image by run mysqlutil/test/dep/rm_imd.sh after test
    dcli = _docker_cli()
    dcli.stop(container=mysql_test_name)



#!/usr/bin/env python2
# coding: utf-8

import logging
import time

logger = logging.getLogger(__name__)


def run(job):

    data = job.data

    for i in range(10):
        data['n'] = i
        time.sleep(1)



#!/usr/bin/env python2
# coding: utf-8

import re
import socket
import unittest
import uuid

import psutil

from pykit import cluster
from pykit import proc
from pykit import ututil

dd = ututil.dd


class TestClusterServer(unittest.TestCase):

    def test_make_server_id(self):
        server_id = str(cluster.ServerID.local_server_id())
        self.assertEqual(12, len(server_id))
        self.assertRegexpMatches(server_id, "^[0-9a-f]{12}$")
        out = proc.shell_script('ifconfig')
        self.assertIn(re.sub('(.{2})', r':\1', server_id)[1:], str(out))

    def test_validate_server_id(self):
        invalid_cases = (
            (),
            [],
            {},
            'foo',
            123,

            'aabbccddeef,'
            'aabbccddeef('
            'aabbccddeef?'
            'aabbccddeef&'
            'aabbccddEEff'
            'aabbccddeeffgg'
            'aabbccddeegg'
        )

        for c in invalid_cases:
            self.assertFalse(cluster.ServerID.validate(c))

        cases = (
            '112233aabbcc',
            'aa112233bbcc',
            '112233445566',
            'aabbccddeeff',
        )

        for c in cases:
            self.assertTrue(cluster.ServerID.validate(c))

    def test_server_id_to_str(self):
        sid = cluster.ServerID.local_server_id()
        self.assertIsInstance(sid, cluster.ServerID)
        self.assertEqual('%012x' % uuid.getnode(), str(sid))
        self.assertEqual('%012x' % uuid.getnode(), sid)

    def test_serverrec(self):
        cases = (
            ('.l1', 'center', {'role1': 1}, {'foo': 1}),
            ('.l1.l2', 'xx', {'role1': 1, 'role2': 1}, {'foo': 1, 'bar': 2}),
            ('.l1.l2.l3', 'yy', {'role1': 1, 'role2': 1, 'role3': 1}, {'foo': 1}),
            ('.l1.l2.l3.l4', 'zz', {'role1': 1, 'role4': 4}, {'foo': 1, 'bar': 2, 'foobar': 3}),
        )

        for idc, idc_type, roles, argkv in cases:
            serverrec = cluster.make_serverrec(idc, idc_type, roles, '/s2', **argkv)

            dd('serverrec:' + repr(serverrec))

            self.assertIn('server_id', serverrec)
            self.assertIn('pub_ips', serverrec)
            self.assertIn('inn_ips', serverrec)
            self.assertIn('memory', serverrec)

            self.assertIn('count', serverrec['cpu'])
            if hasattr(psutil, 'cpu_freq'):
                self.assertIn('frequency', serverrec['cpu'])

            self.assertIn('mountpoints', serverrec)
            self.assertEqual(1, serverrec['next_mount_index'])
            self.assertIn('allocated_drive', serverrec)

            for k, v in serverrec['allocated_drive'].items():
                self.assertIn('/s2', k)
                self.assertEqual({'status': 'normal'}, v)

            self.assertEqual(socket.gethostname(), serverrec['hostname'])
            self.assertEqual(idc, serverrec['idc'])
            self.assertEqual(idc_type, serverrec['idc_type'])
            self.assertEqual(roles, serverrec['roles'])

            for k, v in argkv.items():
                self.assertEqual(v, serverrec[k])

            serverrec_str = cluster.get_serverrec_str(serverrec)
            dd('serverrec_str:' + repr(serverrec_str))

            exp_str = ('server_id: {sid}; idc: {idc}; idc_type: {idct}; roles: {r};'
                       ' mountpoints_count: {mp_cnt}; allocated_drive_count: {ad_cnt}').format(
                sid=serverrec['server_id'], idc=idc, idct=idc_type,
                r=roles, mp_cnt=len(serverrec['mountpoints']),
                ad_cnt=len(serverrec['allocated_drive']))

            self.assertEqual(exp_str, serverrec_str)

    def test_drive_id(self):
        cases = (
            ('112233445566', 1),
            ('112233445566', 10),
            ('112233445566', 100),
            ('112233445566', 999),

            ('aabbccddeeff', 1),
            ('aabbccddeeff', 10),
            ('aabbccddeeff', 100),
            ('aabbccddeeff', 999),

            ('1122ccddeeff', 1),
            ('1122ccddeeff', 10),
            ('1122ccddeeff', 100),
            ('1122ccddeeff', 999),
        )

        for sid, mp_idx in cases:
            drive_id = str(cluster.DriveID(sid, mp_idx))
            self.assertEqual('%s0%03d' % (sid[:12], mp_idx % 1000),
                             drive_id)

            self.assertEqual(sid, cluster.DriveID.parse(drive_id).server_id)
            self.assertEqual(mp_idx, cluster.DriveID.parse(drive_id).mountpoint_index)

    def test_drive_id_server_id(self):

        for drive_id in (cluster.DriveID('112233445566', 1),
                         cluster.DriveID.parse('1122334455660001')):

            dd(drive_id)

            self.assertIsInstance(drive_id.server_id, str)
            self.assertEqual('112233445566', drive_id.server_id)
            self.assertEqual('1122334455660001', drive_id.tostr())
            self.assertEqual('1122334455660001', str(drive_id))

    def test_validate_drive_id(self):
        invalid_cases = (
            (),
            [],
            {},
            'foo',
            123,
            'aabbccddeeff*001',
            'aabbccddeeff000a',
            '*&bbccddeeff0001',
            'AAbbccddeeff0001',
            'AAbbccddeeff0xxx',
        )

        for c in invalid_cases:
            self.assertFalse(cluster.DriveID.validate(c))

        cases = (
            'aabbccddeeff0001',
            'aabbccddeeff0100',
            'aabbccddeeff0999',
            '11bbccddeeff0999',
            '11bb33ddeeff0999',
        )

        for c in cases:
            self.assertTrue(cluster.DriveID.validate(c))

    def test_validate_idc(self):
        invalid_cases = (
            {},
            (),
            [],
            'l1',
            '*l1.l2',
            '&l1.l2.l3',

            '.l&&',
            '.l1.l*',
        )

        for c in invalid_cases:
            self.assertFalse(cluster.validate_idc(c))

        cases = (
            '',
            '.',
            '.l1',
            '.l1.l2',
            '.l1.l2.l3',
            '.l1.l2.l3.l4',

            '.L1',
            '.L1.L2',
            '.L1.L2.L3',
            '.L1.L2.L3.L4',
        )

        for c in cases:
            self.assertTrue(cluster.validate_idc(c))

    def test_idc_distance(self):
        cases = (
            ('', '', 0),
            ('.', '.', 0),
            ('.wt', '.wt', 0),
            ('.wt.TJ', '.wt.TJ', 0),
            ('.wt.TJ.xx', '.wt.TJ.yy', 4),
            ('.wt.XD', '.wt.TJ', 8),
            ('.dx.GZ', '.wt.TJ', 16),
            ('.dx', '.wt', 16),
            ('.dx.GZ', '.dx', 8),
        )

        for n1, n2, dis in cases:
            self.assertEqual(dis,
                             cluster.idc_distance(n1, n2))



#!/usr/bin/env python2
# coding: utf-8

import multiprocessing
import os
import random
import time
import unittest

from pykit import ututil
from pykit.cgrouparch import cgroup_manager
from pykit.cgrouparch import cgroup_util

dd = ututil.dd

random.seed(time.time())


class TestCpu(unittest.TestCase):

    def compute_work(self):
        i = 0
        while True:
            i += 1
            if i > 1024 * 1024:
                break

    def worker(self, index, duration, result_dict):
        # wait for the cgroup directory tree to be setup.
        time.sleep(0.2)

        start_time = time.time()
        dd('worker %d %d started at: %f' %
           (index, os.getpid(), start_time))

        count = 0
        while True:
            self.compute_work()
            count += 1
            dd('worker %d %d computed %d times' %
               (index, os.getpid(), count))

            if time.time() - start_time > duration:
                break

        dd('worker %d %d stoped at: %f' %
           (index, os.getpid(), time.time()))

        result_dict[index] = count
        return

    def test_cpu_share(self):
        manager = multiprocessing.Manager()
        result_dict = manager.dict()

        p1 = multiprocessing.Process(target=self.worker,
                                     args=(1, 10, result_dict))
        p1.daemon = True
        p1.start()

        p2 = multiprocessing.Process(target=self.worker,
                                     args=(2, 10, result_dict))
        p2.daemon = True
        p2.start()

        p3 = multiprocessing.Process(target=self.worker,
                                     args=(3, 10, result_dict))
        p3.daemon = True
        p3.start()

        arch_conf = {
            'cpu': {
                'sub_cgroup': {
                    'test_cgroup_a': {
                        'conf': {
                            'share': 500,
                            'pids': [p1.pid],
                        },
                    },
                    'test_cgroup_b': {
                        'conf': {
                            'share': 1000,
                        },
                        'sub_cgroup': {
                            'test_cgroup_b_sub1': {
                                'conf': {
                                    'share': 100,
                                    'pids': [p2.pid],
                                },
                            },
                            'test_cgroup_b_sub2': {
                                'conf': {
                                    'share': 300,
                                    'pids': [p3.pid],
                                },
                            },
                        },
                    },
                },
            },
        }

        context = {
            'cgroup_dir': '/sys/fs/cgroup',
            'arch_conf': {
                'value': arch_conf
            },
        }

        cgroup_manager.build_all_subsystem_cgroup_arch(context)
        cgroup_manager.set_cgroup(context)

        p1.join()
        p2.join()
        p3.join()

        dd(result_dict)

        for cgrou_name in arch_conf['cpu']['sub_cgroup'].keys():
            cgroup_util.remove_cgroup(
                os.path.join(context['cgroup_dir'], 'cpu'),
                os.path.join(context['cgroup_dir'], 'cpu', cgrou_name))

        level1_rate = float(result_dict[2] + result_dict[3]) / result_dict[1]
        dd(level1_rate)

        self.assertAlmostEqual(2.0, level1_rate, delta=0.2)

        level2_rate = float(result_dict[3]) / result_dict[2]
        dd(level2_rate)

        self.assertAlmostEqual(3.0, level2_rate, delta=0.5)




import gc
import json
import logging
import os
import sys
import threading
import time

import objgraph
import psutil


def start_mem_check_thread(threshold=1024 * 1024 * 1024,
                           gc=False,
                           size_range=None,
                           interval=1
                           ):
    """
    Start a thread in background and in daemon mode, to watch memory usage.
    If memory this process is using beyond `threshold`, a memory usage profile
    is made and is written to root logger. And process is aborted.

    `threshold`:    maximum memory a process can use before abort.
    `gc`:           whether to run gc every time before checking memory usage.
    `size_range`:   in tuple, dump only object of size in this range.
    `interval`:     memory check interval.
    """

    options = {
        'threshold': threshold,
        'gc': gc,
        'size_range': size_range,
        'interval': interval,
    }

    th = threading.Thread(target=mem_check, args=(options,))
    th.daemon = True
    th.start()

    return th


def mem_check(opts):

    while True:

        if opts['gc']:
            try:
                gc.collect()
            except Exception as e:
                logging.exception(repr(e) + ' while gc.collect()')

        try:
            rss = psutil.Process(os.getpid()).memory_info().rss

            logging.info('current memory used: {rss}'.format(rss=rss))

            if rss > opts['threshold']:
                memory_dump(opts)
                os.abort()
        except Exception as e:
            logging.exception(repr(e) + ' while checking memory usage')

        finally:
            time.sleep(opts['interval'])


def memory_dump(opts):
    try:
        _memory_dump(opts)
    except Exception as e:
        logging.exception(repr(e))


def _memory_dump(opts):

    for typ, n in objgraph.most_common_types():
        logging.info('{typ:30} {n:>10}'.format(typ=typ, n=n))

    objects = []
    rng = opts['size_range']

    summ = {
        'max_refsize': {
            'size': 0,
        },
    }

    for obj in gc.get_objects():

        if not hasattr(obj, '__class__'):
            continue

        size = sys.getsizeof(obj, 0)

        if rng is not None:
            if not (rng[0] <= size < rng[1]):
                continue

        i = id(obj)

        # referrers = [id(o)
        #              for o in gc.get_referrers(obj)
        #              if hasattr(o, '__class__')]

        referents = [(id(o), _get_class(o), sys.getsizeof(o, 0))
                     for o in gc.get_referents(obj)
                     # if hasattr(o, '__class__')
                     ]

        refsize = sum([x[2] for x in referents])

        cls = _get_class(obj)

        data = [
            i,
            cls,
            size,           # object size
            refsize,        # size of all direct referents
            referents,      # referents
        ]

        objects.append(data)

        if summ['max_refsize']['size'] < refsize:

            summ['max_refsize'] = {
                'size': refsize,
                'object': data,
            }

    for o in objects:
        logging.info('memory-dump: ' + json.dumps(o))

    logging.info('memory-dump summary: ' + json.dumps(summ))


def _get_class(obj):

    if hasattr(obj, '__class__') and hasattr(obj.__class__, '__name__'):
        cls = obj.__class__.__name__
    else:
        cls = str(type(obj))

    return cls

if __name__ == "__main__":

    logging.basicConfig(level='DEBUG',
                        format='%(asctime)s,%(name)s,%(levelname)s %(message)s',
                        datefmt='%H:%M:%S'
                        )

    rss = psutil.Process(os.getpid()).memory_info().rss

    logging.info('initial mem:' + repr(rss))

    start_mem_check_thread(threshold=rss + 1024 * 100,
                           interval=0.1)

    a = []

    logging.info('a: ' + str(id(a)))

    while True:
        a.append(str(int(time.time() * 1000)) * 100)
        time.sleep(0.001)




from kazoo.exceptions import BadVersionError
from kazoo.exceptions import NoNodeError

from pykit import ututil
from pykit import zktx
from pykit.zktx.test import base

dd = ututil.dd


class TestZKKVAccessor(base.ZKTestBase):

    def test_set_or_create_acl(self):

        kv = zktx.ZKKeyValue(self.zkauthed)

        kv.set_or_create('tacl', '1')

        acl, _ = self.zkauthed.get_acls('tacl')

        self.assertEqual(acl, self.zkauthed._zkconf.kazoo_digest_acl())

    def test_get_path(self):

        kv = zktx.ZKKeyValue(self.zk, get_path=lambda x: 'foo-' + x)

        kv.create('bar', '1')

        rst, ver = self.zk.get('foo-bar')
        self.assertEqual('1', rst)

        rst, ver = kv.get('bar')
        self.assertEqual('1', rst)

        kv.set('bar', '2')
        rst, ver = kv.get('bar')
        self.assertEqual('2', rst)

        kv.delete('bar')
        self.assertRaises(NoNodeError, kv.get, 'bar')

    def test_load(self):

        kv = zktx.ZKKeyValue(self.zk, load=lambda v: '(' + v + ')')

        self.zk.create('foo', '1')
        rst, ver = kv.get('foo')
        self.assertEqual('(1)', rst)

    def test_dump(self):

        kv = zktx.ZKKeyValue(self.zk, dump=lambda v: '(' + v + ')')

        kv.create('foo', '1')
        rst, ver = self.zk.get('foo')
        self.assertEqual('(1)', rst)

        kv.set('foo', '2')
        rst, ver = self.zk.get('foo')
        self.assertEqual('(2)', rst)

    def test_nonode_callback(self):

        kv = zktx.ZKKeyValue(self.zk, nonode_callback=lambda: ('default', 10))

        rst, ver = kv.get('foo')
        self.assertEqual(('default', 10), (rst, ver))

    def test_version(self):

        kv = zktx.ZKKeyValue(self.zk, get_path=lambda x: 'foo-' + x)

        self.zk.create('foo-bar', '1')

        kv.set('bar', '2', version=0)
        rst, ver = kv.get('bar')
        self.assertEqual('2', rst)

        self.assertRaises(BadVersionError, kv.set, 'bar', '2', version=0)
        self.assertRaises(BadVersionError, kv.set, 'bar', '2', version=2)

        self.assertRaises(BadVersionError, kv.delete, 'bar', version=0)
        self.assertRaises(BadVersionError, kv.delete, 'bar', version=2)

        kv.delete('bar')

    def test_set_or_create(self):

        kv = zktx.ZKKeyValue(self.zk)

        # create

        kv.set_or_create('foo', '1')

        rst, ver = kv.get('foo')
        self.assertEqual('1', rst)
        self.assertEqual(0, ver)

        # set

        self.assertRaises(BadVersionError, kv.set_or_create, 'foo', '2', version=2)

        kv.set_or_create('foo', '2', version=0)
        kv.set_or_create('foo', '2', version=1)
        rst, ver = kv.get('foo')
        self.assertEqual('2', rst)
        self.assertEqual(2, ver)


class TestZKValueAccessor(base.ZKTestBase):

    def test_get_path(self):

        v = zktx.ZKValue(self.zk, get_path=lambda: 'foopath')

        v.create('1')

        rst, ver = self.zk.get('foopath')
        self.assertEqual('1', rst)

        rst, ver = v.get()
        self.assertEqual('1', rst)

        v.set('2')
        rst, ver = v.get()
        self.assertEqual('2', rst)

        v.delete()
        self.assertRaises(NoNodeError, v.get)

    def test_load(self):

        v = zktx.ZKValue(self.zk, get_path=lambda: 'foopath',
                         load=lambda v: '(' + v + ')')

        self.zk.create('foopath', '1')
        rst, ver = v.get()
        self.assertEqual('(1)', rst)

    def test_dump(self):

        v = zktx.ZKValue(self.zk, get_path=lambda: 'foopath',
                         dump=lambda v: '(' + v + ')')

        v.create('1')
        rst, ver = self.zk.get('foopath')
        self.assertEqual('(1)', rst)

        v.set('2')
        rst, ver = self.zk.get('foopath')
        self.assertEqual('(2)', rst)

    def test_nonode_callback(self):

        v = zktx.ZKValue(self.zk, get_path=lambda: 'foopath',
                         nonode_callback=lambda: ('default', 10))

        rst, ver = v.get()
        self.assertEqual(('default', 10), (rst, ver))

    def test_version(self):

        v = zktx.ZKValue(self.zk, get_path=lambda: 'foopath')

        self.zk.create('foopath', '1')

        v.set('2', version=0)
        rst, ver = v.get()
        self.assertEqual('2', rst)

        self.assertRaises(BadVersionError, v.set, '2', version=0)
        self.assertRaises(BadVersionError, v.set, '2', version=2)

        self.assertRaises(BadVersionError, v.delete, version=0)
        self.assertRaises(BadVersionError, v.delete, version=2)

        v.delete()

    def test_set_or_create(self):

        v = zktx.ZKValue(self.zk, get_path=lambda: 'foopath')

        # create

        v.set_or_create('1')

        rst, ver = v.get()
        self.assertEqual('1', rst)
        self.assertEqual(0, ver)

        # set

        self.assertRaises(BadVersionError, v.set_or_create, '2', version=2)

        v.set_or_create('2', version=0)
        v.set_or_create('2', version=1)
        rst, ver = v.get()
        self.assertEqual('2', rst)
        self.assertEqual(2, ver)

    def test_set_or_create_acl(self):
        v = zktx.ZKValue(self.zkauthed, get_path=lambda: 'taclvalue')

        # create

        v.set_or_create('1')

        acl, _ = self.zkauthed.get_acls('taclvalue')
        self.assertEqual(acl, self.zkauthed._zkconf.kazoo_digest_acl())



#!/usr/bin/env python2
# coding: utf-8

import calendar
import datetime
import time
import types

import pytz
import tzlocal

formats = {
    'default':        '%a, %d %b %Y %H:%M:%S UTC',
    'utc':            '%a, %d %b %Y %H:%M:%S UTC',
    'iso':            '%Y-%m-%dT%H:%M:%S.000Z',
    'archive':        '%Y%m%d-%H',
    'compact':        '%Y%m%d-%H%M%S',
    'daily':          '%Y-%m-%d',
    'daily_compact':  '%Y%m%d',
    'mysql':          '%Y-%m-%d %H:%M:%S',
    'nginxaccesslog': "%d/%b/%Y:%H:%M:%S",
    'nginxerrorlog':  "%Y/%m/%d %H:%M:%S",
}

ts_length = {
    's':  10,
    'ms': 13,
    'us': 16,
    'ns': 19,
}


def parse(time_str, fmt_key, timezone=None):
    dt = datetime.datetime.strptime(time_str, _get_format(fmt_key))
    if timezone is not None:
        tz = pytz.timezone(timezone)
        dt = tz.localize(dt)

    return dt


def format(dt, fmt_key):
    return dt.strftime(_get_format(fmt_key))


def format_ts(ts, fmt_key, utc=True):
    dt = ts_to_datetime(ts, utc)
    return format(dt, fmt_key)


def _get_format(fmt_key):
    return formats.get(fmt_key) or fmt_key


def utc_datetime_to_ts(dt):
    return calendar.timegm(dt.timetuple())


def datetime_to_ts(dt):
    epoch_dt = datetime.datetime.fromtimestamp(0, tz=pytz.utc)

    if not hasattr(dt, 'tzinfo') or dt.tzinfo is None:
        local_tz = tzlocal.get_localzone()
        dt = local_tz.localize(dt)

    delta = dt - epoch_dt
    ts = delta.total_seconds()

    return ts


def ts_to_datetime(ts, utc=True):
    if utc:
        return datetime.datetime.utcfromtimestamp(ts)
    else:
        return datetime.datetime.fromtimestamp(ts)


def ts():
    return int(time.time())


def ms():
    return int(time.time() * 1000)


def us():
    return int(time.time() * (1000 ** 2))


def ns():
    return int(time.time() * (1000 ** 3))


def ms_to_ts(ms):
    return ms / 1000


def us_to_ts(us):
    return us / (1000 ** 2)


def ns_to_ts(ns):
    return ns / (1000 ** 3)


def to_sec(v):
    """
    Convert millisecond, microsecond or nanosecond to second.

    ms_to_ts, us_to_ts, ns_to_ts are then deprecated.
    """

    v = float(str(v))

    if (type(v) != types.FloatType
            or v < 0):
        raise ValueError('invalid time to convert to second: {v}'.format(v=v))

    l = len(str(int(v)))

    if l == 10:
        return int(v)
    elif l == 13:
        return int(v / 1000)
    elif l == 16:
        return int(v / (1000**2))
    elif l == 19:
        return int(v / (1000**3))
    else:
        raise ValueError(
            'invalid time length, not 10, 13, 16 or 19: {v}'.format(v=v))


def is_timestamp(string, unit=None):

    string = str(string)

    if not string.isdigit():
        return False

    if unit is None:
        return len(string) in ts_length.values()

    if unit in ts_length:
        return len(string) == ts_length[unit]

    return False



import errno
import fcntl
import logging
import os
import signal
import sys
import time

import __main__

logger = logging.getLogger(__name__)


class Daemon(object):

    def __init__(self,
                 pidfile=None,
                 stdin='/dev/null',
                 stdout='/dev/null',
                 stderr='/dev/null',
                 close_fds=False):

        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pidfile = pidfile or _default_pid_file()
        # NOTE: We need to open another separate file to avoid the file
        #       being reopened again.
        #       In which case, process loses file lock.
        #
        # From "man fcntl":
        # As well as being removed by an explicit F_UNLCK, record locks are
        # automatically released when the process terminates or if it
        # closes any file descriptor referring to a file on which locks
        # are held. This is bad: it means that a process can lose the locks
        # on a file like /etc/passwd or /etc/mtab when for some reason a
        # library function decides to open, read and close it.
        self.lockfile = self.pidfile + ".lock"
        self.lockfp = None
        self.close_fds = close_fds

    def daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """

        try:

            pid = os.fork()
            if pid > 0:
                # exit first parent
                _close_std_io()
                sys.exit(0)

        except OSError as e:
            logger.error("fork #1 failed: " + repr(e))
            sys.exit(1)

        # decouple from parent environment
        os.setsid()
        os.umask(0)

        # do second fork
        try:

            pid = os.fork()
            if pid > 0:
                # exit from second parent
                _close_std_io()
                sys.exit(0)

        except OSError as e:
            logger.error("fork #2 failed: " + repr(e))
            sys.exit(1)

        if self.close_fds:
            _close_fds()

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = file(self.stdin, 'r')
        so = file(self.stdout, 'a+')
        se = file(self.stderr, 'a+', 0)
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        logger.info("OK daemonized")

    def trylock_or_exit(self, timeout=10):

        interval = 0.1
        n = int(timeout / interval) + 1
        flag = fcntl.LOCK_EX | fcntl.LOCK_NB

        for ii in range(n):

            fd = os.open(self.lockfile, os.O_RDWR | os.O_CREAT)

            fcntl.fcntl(fd, fcntl.F_SETFD,
                        fcntl.fcntl(fd, fcntl.F_GETFD, 0)
                        | fcntl.FD_CLOEXEC)

            try:
                fcntl.lockf(fd, flag)

                self.lockfp = os.fdopen(fd, 'w+r')
                break

            except IOError as e:
                os.close(fd)
                if e[0] == errno.EAGAIN:
                    time.sleep(interval)
                else:
                    raise

        else:
            logger.info("Failure acquiring lock %s" % (self.lockfile, ))
            sys.exit(1)

        logger.info("OK acquired lock %s" % (self.lockfile))

    def unlock(self):

        if self.lockfp is None:
            return

        fd = self.lockfp.fileno()
        fcntl.lockf(fd, fcntl.LOCK_UN)
        self.lockfp.close()
        self.lockfp = None

    def start(self):

        self.daemonize()
        self.init_proc()

    def init_proc(self):
        self.trylock_or_exit()
        self.write_pid_or_exit()

    def write_pid_or_exit(self):

        self.pf = open(self.pidfile, 'w+r')
        pf = self.pf

        fd = pf.fileno()
        fcntl.fcntl(fd, fcntl.F_SETFD,
                    fcntl.fcntl(fd, fcntl.F_GETFD, 0)
                    | fcntl.FD_CLOEXEC)

        try:
            pid = os.getpid()
            logger.debug('write pid:' + str(pid))

            pf.truncate(0)
            pf.write(str(pid))
            pf.flush()
        except Exception as e:
            logger.exception('write pid failed.' + repr(e))
            sys.exit(0)

    def stop(self):

        pid = None

        if not os.path.exists(self.pidfile):

            logger.debug('pidfile not exist:' + self.pidfile)
            return

        try:
            pid = _read_file(self.pidfile)
            pid = int(pid)
            os.kill(pid, signal.SIGTERM)
            return

        except Exception as e:
            logger.warn('{e} while get and kill pid={pid}'.format(
                e=repr(e), pid=pid))


def _read_file(fn):
    with open(fn, 'r') as f:
        return f.read()


def _close_std_io():
    os.close(0)
    os.close(1)
    os.close(2)


def _close_fds():

    try:
        max_fd = os.sysconf("SC_OPEN_MAX")
    except ValueError as e:
        logger.warn(repr(e) + ' while get max fds of a process')
        max_fd = 65536

    for i in xrange(3, max_fd):
        try:
            os.close(i)
        except OSError:
            pass


def _default_pid_file():

    if hasattr(__main__, '__file__'):
        name = __main__.__file__
        name = os.path.basename(name)
        if name == '<stdin>':
            name = '__stdin__'
        return '/var/run/' + name.rsplit('.', 1)[0]
    else:
        return '/var/run/pykit.daemonize'


def daemonize_cli(run_func, pidfn, close_fds=False):

    logging.basicConfig(stream=sys.stderr)
    logging.getLogger(__name__).setLevel(logging.DEBUG)

    d = Daemon(pidfile=pidfn, close_fds=close_fds)

    logger.info("sys.argv: " + repr(sys.argv))

    try:
        if len(sys.argv) == 1:
            d.init_proc()
            run_func()

        elif len(sys.argv) == 2:

            if 'start' == sys.argv[1]:
                d.start()
                run_func()

            elif 'stop' == sys.argv[1]:
                d.stop()

            elif 'restart' == sys.argv[1]:
                d.stop()
                d.start()
                run_func()

            else:
                logger.error("Unknown command: %s" % (sys.argv[1]))
                print "Unknown command"
                sys.exit(2)

            sys.exit(0)
        else:
            print "usage: %s start|stop|restart" % sys.argv[0]
            sys.exit(2)

    except Exception as e:
        logger.exception(repr(e))


standard_daemonize = daemonize_cli



from .jobscheduler import (
    JobExistError,
    JobScheduler,
    NextFireTimeError,
    get_next_fire_time,
)

__all__ = [
    'JobExistError',
    'JobScheduler',
    'NextFireTimeError',
    'get_next_fire_time',
]



import logging
import time

logger = logging.getLogger(__name__)


def limit_job_speed(max_job_speed, job_step=1):
    speed_stat = {
        'start_time': time.time(),
        'job_num': 0,
        'tick_time': 5,
    }

    def _limit(job_args):
        try:
            max_speed = max_job_speed
            if callable(max_job_speed):
                max_speed = max_job_speed()

            speed_stat['job_num'] += job_step

            min_ts = speed_stat['job_num'] * 1.0 / max_speed
            itv_ts = time.time() - speed_stat['start_time']
            if itv_ts < min_ts:
                time.sleep(min_ts - itv_ts)

            now = time.time()
            if now - speed_stat['start_time'] >= speed_stat['tick_time']:
                speed = speed_stat['job_num'] * 1.0 / (now - speed_stat['start_time'])

                logger.info('current speed: {speed}/s max_speed:{max_speed}/s'
                            ' job_num:{job_num} start_time:{start_time}'.format(
                                speed=round(speed, 3),
                                max_speed=max_speed,
                                job_num=speed_stat['job_num'],
                                start_time=speed_stat['start_time']))

                speed_stat['start_time'] = now
                speed_stat['job_num'] = 0

        except Exception as e:
            logger.exception('error occurred limit job speed: ' + repr(e))

        return job_args

    return _limit



import unittest

from kazoo.client import KazooClient

from pykit import threadutil
from pykit import utdocker
from pykit import ututil
from pykit import zkutil

dd = ututil.dd

zk_tag = 'daocloud.io/zookeeper:3.4.10'
zk_name = 'zk_test'


class TestAcid(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        utdocker.pull_image(zk_tag)

    def setUp(self):

        utdocker.create_network()
        utdocker.start_container(
            zk_name,
            zk_tag,
            env={
                "ZOO_MY_ID": 1,
                "ZOO_SERVERS": "server.1=0.0.0.0:2888:3888",
            },
            port_bindings={2181: 2181}
        )

        self.zk = KazooClient('127.0.0.1:2181')
        self.zk.start()

        dd('start zk-test in docker')

        self.path = 'a'
        self.zk.create(self.path, '1')

    def tearDown(self):

        self.zk.stop()
        utdocker.remove_container(zk_name)

    def test_cas(self):

        for curr in zkutil.cas_loop(self.zk, self.path):
            curr.v += 2

        final_val, zstat = self.zk.get(self.path)
        dd(final_val, zstat)
        self.assertEqual('3', final_val)

    def test_cas_abort(self):

        for curr in zkutil.cas_loop(self.zk, self.path):
            curr.v += 2
            break

        final_val, zstat = self.zk.get(self.path)
        dd(final_val, zstat)
        self.assertEqual('1', final_val, 'a break statement cancel set_val()')

    def test_cas_create_zk(self):

        for first_arg in (
                '127.0.0.1:2181',
                {'hosts': '127.0.0.1:2181'},
                zkutil.ZKConf(hosts='127.0.0.1:2181'),
        ):
            self.zk.set(self.path, '1')

            for curr in zkutil.cas_loop(first_arg, self.path):
                curr.v += 2

            final_val, zstat = self.zk.get(self.path)
            dd(final_val, zstat)
            self.assertEqual('3', final_val)

    def test_cas_non_json(self):

        for curr in zkutil.cas_loop(self.zk, self.path, json=False):
            self.assertIsInstance(curr.v, str)
            curr.v += 'a'

        final_val, zstat = self.zk.get(self.path)
        dd(final_val, zstat)
        self.assertEqual('1a', final_val)

    def test_cas_concurrent(self):

        def _update():
            for ii in range(10):
                for curr in zkutil.cas_loop('127.0.0.1:2181', self.path):
                    curr.v += 1

        ths = [threadutil.start_daemon(_update)
               for _ in range(5)]

        for th in ths:
            th.join()

        final_val, zstat = self.zk.get(self.path)
        dd(final_val, zstat)

        self.assertEqual('51', final_val)



#!/usr/bin/env python2
# coding: utf-8

import argparse
import copy
import sys
import logging
import os

from pykit import dictutil


logger = logging.getLogger(__name__)

def command(**kwargs):

    root, parser = add_command_help(kwargs)

    inputs = sys.argv[1:]

    try:
        cmds = []
        while len(inputs) > 0 and root.has_key(inputs[0]):

            k = inputs.pop(0)
            cmds.append(k)
            node = root[k]

            if is_node_executable(node):

                call_able, args = parse_executable_node(parser, cmds, node, inputs)

                try:
                    logger.debug("command: " + repr(cmds) + ' args: ' + repr(args) + ' cwd: ' + repr(os.getcwd()))

                    rc = call_able(*args)
                    sys.exit(0
                              if rc is True or rc is 0 or rc is None
                              else 1)

                except Exception as e:
                    logger.exception(repr(e))
                    sys.stderr.write(repr(e))
                    sys.exit(1)

            else:
                root = node

        if need_to_show_help(parser):

            if len(cmds) > 0:
                argv = [' '.join(cmds)] + inputs
            else:
                argv = inputs

            parser.parse_args(argv)
        else:
            sys.stderr.write('No such command: ' + ' '.join(sys.argv[1:]))

        sys.exit(2)

    except Exception as e:
        logger.exception(repr(e))
        sys.stderr.write(repr(e))
        sys.exit(1)


def add_command_help(commands):

    new_cmds = copy.deepcopy(commands)

    help_msgs = new_cmds.get('__add_help__')
    desc = new_cmds.get('__description__')

    for k in ('__add_help__', '__description__'):
        if new_cmds.has_key(k):
            del new_cmds[k]

    if help_msgs is None:
        return new_cmds, None

    parser = argparse.ArgumentParser(description=desc, epilog='\n')

    subparsers = parser.add_subparsers(help=' command(s) to select ...')

    for cmds, execute_able in dictutil.depth_iter(new_cmds):

        help = help_msgs.get(tuple(cmds), '')
        cmd = ' '.join(cmds)

        cmd_parser = subparsers.add_parser(cmd, help=help)

        if need_param_help(execute_able):
            call_able = execute_able[0]
            param_msgs = execute_able[1:]

            params = add_param_help(cmd_parser, param_msgs)

            # delete help message
            dictutil.make_setter(cmds)(new_cmds, (call_able, params))

    return new_cmds, parser


def add_param_help(parser, param_msgs):

    params = []
    for param, msg in param_msgs:
        parser.add_argument(param, **msg)
        params.append(param)

    return params


def parse_executable_node(parser, cmds, execute_able, args):

    if not need_to_show_help(parser):

        # no __add_help__ but has paramter help message
        if args_need_to_parse(execute_able):
            return execute_able[0], args

        return execute_able, args

    args_parsed = parser.parse_args([' '.join(cmds)] + args)
    # to dict
    args_parsed = vars(args_parsed)

    if not args_need_to_parse(execute_able):
        return execute_able, args

    call_able, params = execute_able

    args = [args_parsed.get(x) for x in params]

    return call_able, args


def is_node_executable(node):

    if isinstance(node, (list, tuple)) and len(node) > 0:
        return callable(node[0])

    return callable(node)


def need_to_show_help(parser):

    return parser is not None


def args_need_to_parse(execute_able):

    return isinstance(execute_able, tuple)


def need_param_help(execute_able):

    return isinstance(execute_able, (list, tuple)) and len(execute_able) > 1



#!/usr/bin/env python2
# coding: utf-8

import unittest

from pykit import cluster
from pykit import ututil

dd = ututil.dd


class TestBlockIndex(unittest.TestCase):

    def test_parse_invalid(self):
        cases = (
            '',
            'a',
            '0',
            '00',
            '000',
            '000a',
            'a000',
        )

        for bad in cases:
            self.assertRaises(cluster.BlockIndexError, cluster.BlockIndex.parse, bad)

    def test_parse(self):

        cases = (
                ('0000', (0, 0)),
                ('0001', (0, 1)),
                ('0101', (1, 1)),
                ('1234', (12, 34)),
        )

        for inp, expected in cases:
            dd(inp)
            dd(expected)

            rst = cluster.BlockIndex.parse(inp)
            self.assertEqual(expected, rst)

    def test_new_invalid(self):

        cases = (
            (-1, 0),
            (100, 0),
            (0, -1),
            (0, 100),
        )

        for i, j in cases:
            dd(i, j)
            self.assertRaises(cluster.BlockIndexError, cluster.BlockIndex, i, j)



#!/usr/bin/env python2
# coding: utf-8

import errno
import logging
import select
import socket

from pykit import stopwatch

logger = logging.getLogger(__name__)


class HttpError(Exception):
    pass


class LineTooLongError(HttpError):
    pass


class ChunkedSizeError(HttpError):
    pass


class NotConnectedError(HttpError):
    pass


class ResponseNotReadyError(HttpError):
    pass


class HeadersError(HttpError):
    pass


class BadStatusLineError(HttpError):
    pass


MAX_LINE_LENGTH = 65536
LINE_RECV_LENGTH = 1024 * 4


class Client(object):

    stopwatch_root_name = 'pykit.http.Client'

    def __init__(self, host, port, timeout=60,
                 stopwatch_kwargs=None, https_context=None):

        self.host = host
        self.port = port
        self.timeout = timeout
        self.sock = None
        self.chunked = False
        self.chunk_left = None
        self.content_length = None
        self.has_read = 0
        self.status = None
        self.headers = {}
        self.recv_iter = None
        self.https_context = https_context
        self.request_chunked_encoded = False

        self.stopwatch_kwargs = {
            # min_tracing_milliseconds=0 to trace all events. StopWatch trace
            # only event those cost less than min_tracing_milliseconds
            'min_tracing_milliseconds': 0,
        }

        if stopwatch_kwargs is not None:
            self.stopwatch_kwargs.update(stopwatch_kwargs)

        self.stopwatch = stopwatch.StopWatch(**self.stopwatch_kwargs)
        self.stopwatch_started = False

    def get_trace_str(self):
        tr = self.get_trace()
        rst = []
        for t in tr:
            ent = '{name}: {time:.6f} {annotation}'.format(
                name=t['name'],
                time=float('{:.6f}'.format(t['time'])),
                annotation=','.join(t['annotation']) or '-'
            )
            rst.append(ent)

        return '; '.join(rst)

    def get_trace(self):
        sw = self.get_and_end_stopwatch()
        rst = []
        for t in sw.get_last_trace_report():
            ent = dict(
                name=t.name,
                time=t.end_time - t.start_time,
                annotation=[self._human_annotation(
                    an) for an in t.trace_annotations],
            )

            rst.append(ent)

        return rst

    def _human_annotation(self, annotation):
        an = annotation
        return '{key}:{value}'.format(key=an.key, value=an.value)

    def get_and_end_stopwatch(self):
        '''stopwatch must be stopped before it can be read
        '''
        self.end_stopwatch()
        return self.stopwatch

    def end_stopwatch(self):

        if self.stopwatch_started:
            self.stopwatch.end(self.stopwatch_root_name)
            self.stopwatch_started = False

    def start_stopwatch(self):

        if self.stopwatch_started:
            return

        self.stopwatch.start(self.stopwatch_root_name)
        self.stopwatch_started = True

    def __del__(self):

        self._close()

    def request(self, uri, method='GET', headers={}):

        self.send_request(uri, method=method, headers=headers)

        self.read_response()

    def send_request(self, uri, method='GET', headers={}):

        self._reset_request()
        self.method = method

        self.start_stopwatch()

        with self.stopwatch.timer('conn'):
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.settimeout(self.timeout)
            self.sock.connect((self.host, self.port))
            if self.https_context is not None:
                self.sock = self.https_context.wrap_socket(self.sock,
                                                           server_hostname=self.host)

        bufs = ['{method} {uri} HTTP/1.1'.format(method=method, uri=uri), ]

        headers = headers or {}
        if 'Host' not in headers and 'host' not in headers:
            headers['Host'] = self.host

        if headers.get('Transfer-Encoding') == 'chunked':
            self.request_chunked_encoded = True

        for k, v in headers.items():
            bufs.append('%s: %s' % (k, v))

        bufs.extend(['', ''])

        with self.stopwatch.timer('send_header'):
            self.sock.sendall('\r\n'.join(bufs))

    def send_body(self, body):

        if self.sock is None:
            raise NotConnectedError('socket object is None')

        with self.stopwatch.timer('send_body'):
            if self.request_chunked_encoded:
                body = '{0:x}\r\n{1}\r\n\r\n'.format(len(body), body)

            self.sock.sendall(body)

    def read_status(self, skip_100=True):

        if self.status is not None or self.sock is None:
            raise ResponseNotReadyError('response is unavailable')

        if self.recv_iter is None:
            self.recv_iter = _recv_loop(self.sock, self.timeout)
            self.recv_iter.next()

        # read until we get a non-100 response
        while True:

            status = self._get_response_status()
            if status >= 200:
                break

            # skip the header from the 100 response
            with self.stopwatch.timer('recv_skip_header'):

                while True:

                    skip = self._readline()
                    if skip.strip() == '':
                        break

            if skip_100 is False:
                return status

        self.status = status
        return status

    def read_headers(self):

        with self.stopwatch.timer('recv_header'):

            while True:

                line = self._readline()
                if line == '':
                    break

                kv = line.strip().split(':', 1)
                if len(kv) < 2:
                    raise HeadersError(
                        'invalid headers param line:%s' % (line))
                self.headers[kv[0].lower()] = kv[1].strip()

        if self.status in (204, 304) or self.method == 'HEAD':
            self.content_length = 0
            return self.headers

        code = self.headers.get('transfer-encoding', '')
        if code.lower() == 'chunked':
            self.chunked = True
            return self.headers

        length = self.headers.get('content-length', '0')

        try:
            self.content_length = int(length)
        except ValueError as e:
            logger.error(
                repr(e) + ' while get content-length length:{l}'.format(l=length))
            raise HeadersError('invalid content-length')

        return self.headers

    def read_response(self):

        self.read_status()
        self.read_headers()

        return self.status, self.headers

    def read_body(self, size):

        with self.stopwatch.timer('recv_body'):
            return self._read_body(size)

    def readlines(self, delimiter=None):

        if delimiter is None:
            delimiter = '\n'

        buf = ''
        while True:

            tmp = self._read_body(MAX_LINE_LENGTH)

            if tmp == '':
                if buf != '':
                    yield buf
                break

            buf += tmp

            lines = buf.split(delimiter)
            buf = lines.pop()

            for l in lines:
                yield l + delimiter

    def _read_body(self, size):

        if size is not None and size <= 0:
            return ''

        if self.chunked:
            buf = self._read_chunked(size)
            self.has_read += len(buf)
            return buf

        if size is None:
            size = self.content_length - self.has_read
        else:
            size = min(size, self.content_length - self.has_read)

        if size <= 0:
            return ''

        buf = self._read(size)
        self.has_read += size

        return buf

    def _close(self):

        if self.recv_iter is not None:
            try:
                self.recv_iter.close()
            except Exception as e:
                logger.exception(repr(e) + ' while close recv_iter')

        self.recv_iter = None

        if self.sock is not None:
            try:
                self.sock.close()
            except Exception as e:
                logger.exception(repr(e) + ' while close sock')

        self.sock = None

    def _reset_request(self):

        self._close()
        self.chunked = False
        self.chunk_left = None
        self.content_length = None
        self.has_read = 0
        self.status = None
        self.headers = {}
        self.request_chunked_encoded = False

    def _read(self, size):
        return self.recv_iter.send(('block', size))

    def _readline(self):
        return self.recv_iter.send(('line', None))

    def _get_response_status(self):

        with self.stopwatch.timer('recv_status'):
            line = self._readline()

        vals = line.split(None, 2)
        if len(vals) < 2:
            raise BadStatusLineError('invalid status line:{l}'.format(l=line))

        ver, status = vals[0], vals[1]

        try:
            status = int(status)
        except ValueError as e:
            logger.error(repr(e) + ' while get response status')
            raise BadStatusLineError('status is not int:{l}'.format(l=line))

        if not ver.startswith('HTTP/') or status < 100 or status > 999:
            raise BadStatusLineError('invalid status line:{l}'.format(l=line))

        return status

    def _get_chunk_size(self):

        line = self._readline()

        i = line.find(';')
        if i >= 0:
            # strip chunk-extensions
            line = line[:i]

        try:
            chunk_size = int(line, 16)
        except ValueError as e:
            logger.error(
                repr(e) + ' while get chunk size line:{l}'.format(l=line))
            raise ChunkedSizeError('invalid chunk size')

        return chunk_size

    def _read_chunked(self, size):

        buf = []

        if self.chunk_left == 0:
            return ''

        while size is None or size > 0:

            if self.chunk_left is None:
                self.chunk_left = self._get_chunk_size()

            if self.chunk_left == 0:
                break

            if size is not None:
                read_size = min(size, self.chunk_left)
                size -= read_size
            else:
                read_size = self.chunk_left

            buf.append(self._read(read_size))
            self.chunk_left -= read_size

            if self.chunk_left == 0:
                self._read(len('\r\n'))
                self.chunk_left = None

        if self.chunk_left == 0:

            while True:
                line = self._readline()
                if line == '':
                    break

        return ''.join(buf)


def _recv_loop(sock, timeout):

    bufs = ['']
    mode, size = yield

    while True:

        if mode == 'line':
            buf = bufs[0]
            if '\r\n' in buf:
                rst, buf = buf.split('\r\n', 1)
                bufs[0] = buf
                mode, size = yield rst
                continue
            else:
                if len(buf) >= MAX_LINE_LENGTH:
                    raise LineTooLongError(
                        'line length greater than max_len:{l}'.format(l=len(buf)))
                else:
                    buf += _recv(sock, timeout, LINE_RECV_LENGTH)
                    bufs[0] = buf
                    continue
        else:
            total = len(bufs[0])
            while total < size:
                bufs.append(_recv(sock, timeout, size - total))
                total += len(bufs[-1])

            rst = ''.join(bufs)
            if size < len(rst):
                bufs = [rst[size:]]
                rst = rst[:size]
            else:
                bufs = ['']
            mode, size = yield rst


def _recv(sock, timeout, size):

    buf = ''
    for _ in range(2):
        try:
            buf = sock.recv(size)
            break
        except socket.error as e:
            if len(e.args) <= 0 or e.args[0] != errno.EAGAIN:
                raise

            evin, evout, everr = select.select(
                [sock.fileno()], [], [], timeout)

            if len(evin) <= 0:
                raise socket.timeout(
                    '{second}s timeout while recv'.format(second=timeout))

    if len(buf) <= 0:
        raise socket.error('got empty when recv {l} bytes'.format(l=size))

    return buf



from .threadutil import (
    AsyncRaiseError,
    InvalidThreadIdError,
    raise_in_thread,
    start_thread,
    start_daemon,
    start_daemon_thread,  # deprecated
)


__all__ = [
    'AsyncRaiseError',
    'InvalidThreadIdError',
    'raise_in_thread',
    'start_thread',
    'start_daemon',
    'start_daemon_thread',  # deprecated
]



from .timeutil import (
    formats,
    parse,
    format,
    format_ts,
    utc_datetime_to_ts,
    datetime_to_ts,
    ts_to_datetime,
    ts,
    ms,
    us,
    ns,
    ms_to_ts,
    us_to_ts,
    ns_to_ts,
    to_sec,
    is_timestamp,
)

__all__ = [
    'formats',
    'parse',
    'format',
    'format_ts',
    'utc_datetime_to_ts',
    'datetime_to_ts',
    'ts_to_datetime',
    'ts',
    'ms',
    'us',
    'ns',
    'ms_to_ts',
    'us_to_ts',
    'ns_to_ts',
    'to_sec',
    'is_timestamp',
]



import logging
import Queue
import threading
import time
from datetime import datetime

from pykit import threadutil
from pykit.logcollector import cache_flusher
from pykit.logcollector import scanner
from pykit.logcollector import sender

logger = logging.getLogger(__name__)


def run(**kwargs):
    context = {
        'node_id': kwargs['node_id'],
        'node_ip': kwargs['node_ip'],
        'send_log': kwargs['send_log'],

        'conf': kwargs['conf'],

        'cache_lock': threading.RLock(),

        'cache': {},
        'stat': {},

        'queue': Queue.Queue(1024 * 10),
    }

    # strptime not thread safe, need to call it manually before
    # initiating any thread
    datetime.strptime("2011-04-05", "%Y-%m-%d")

    for log_name in context['conf'].keys():
        context['cache'][log_name] = {}
        context['stat'][log_name] = {}

        threadutil.start_daemon_thread(
            scanner.scan,
            args=(context, log_name))

    threadutil.start_daemon_thread(cache_flusher.run, args=(context,))

    threadutil.start_daemon_thread(sender.run, args=(context,))

    while True:
        # actually it is not an error log, but normally we only report
        # error log, and we want to report this log even it is not
        # an error log.
        logger.error('stat: %s' % context['stat'])

        time.sleep(100)



#!/usr/bin/env python
# coding: utf-8

import logging

from pykit import config
from pykit import rangeset
from pykit import utfjson

from .redisaccessor import RedisKeyValue
from .redisaccessor import RedisValue
from .status import COMMITTED
from .status import STATUS
from .storage import Storage

logger = logging.getLogger(__name__)


class RedisStorage(Storage):

    def __init__(self, redis_cli, txidset_path):
        self.redis_cli = redis_cli
        self.txidset_path = txidset_path

        self.txidset = RedisValue(self.redis_cli,
                                  get_path=self._txidset_path,
                                  load=txidset_load)
        self.record = RedisKeyValue(self.redis_cli)

    def _txidset_path(self):
        return self.txidset_path or getattr(config, 'redis_txidset_path')

    def apply_jour(self, jour):
        for k, v in jour.items():
            self.apply_record(k, v)

    def apply_record(self, key, val):
        # meta/server/<server_id>
        # meta/region/<region_id>
        k_parts = key.split('/', 2)
        if len(k_parts) != 3:
            raise KeyError('invalid key: {k}'.format(k=key))

        if k_parts[0] != 'meta':
            logger.info('{k} no need to save'.format(k=key))
            return

        hashname = k_parts[1]
        hashkey = k_parts[2]

        self.record.hset(hashname, hashkey, val)

    def add_to_txidset(self, status, txid):
        if status not in STATUS:
            raise KeyError('invalid status: ' + repr(status))

        txidset = self.txidset.get()
        txidset[status].add([txid, txid + 1])

        self.txidset.set(txidset)

    def set_txidset(self, status, txidset):
        val = {
            status: txidset,
        }
        self.txidset.set(val)


def txidset_load(value):
    val = utfjson.load(value)
    if val is None:
        val = {}

    committed = val.get(COMMITTED) or []

    rst = {}
    rst[COMMITTED] = rangeset.RangeSet(committed)

    return rst
